# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: a3m/api/transferservice/v1beta1/request_response.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from google.protobuf import timestamp_pb2 as google_dot_protobuf_dot_timestamp__pb2


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n6a3m/api/transferservice/v1beta1/request_response.proto\x12\x1f\x61\x33m.api.transferservice.v1beta1\x1a\x1fgoogle/protobuf/timestamp.proto"\x80\x01\n\rSubmitRequest\x12\x12\n\x04name\x18\x01 \x01(\tR\x04name\x12\x10\n\x03url\x18\x02 \x01(\tR\x03url\x12I\n\x06\x63onfig\x18\x03 \x01(\x0b\x32\x31.a3m.api.transferservice.v1beta1.ProcessingConfigR\x06\x63onfig" \n\x0eSubmitResponse\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id"\x1d\n\x0bReadRequest\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id"\xa2\x01\n\x0cReadResponse\x12\x46\n\x06status\x18\x01 \x01(\x0e\x32..a3m.api.transferservice.v1beta1.PackageStatusR\x06status\x12\x10\n\x03job\x18\x02 \x01(\tR\x03job\x12\x38\n\x04jobs\x18\x03 \x03(\x0b\x32$.a3m.api.transferservice.v1beta1.JobR\x04jobs")\n\x10ListTasksRequest\x12\x15\n\x06job_id\x18\x01 \x01(\tR\x05jobId"P\n\x11ListTasksResponse\x12;\n\x05tasks\x18\x01 \x03(\x0b\x32%.a3m.api.transferservice.v1beta1.TaskR\x05tasks"\x0e\n\x0c\x45mptyRequest"\x0f\n\rEmptyResponse"\xb9\x02\n\x03Job\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x12\n\x04name\x18\x02 \x01(\tR\x04name\x12\x14\n\x05group\x18\x03 \x01(\tR\x05group\x12\x17\n\x07link_id\x18\x04 \x01(\tR\x06linkId\x12\x43\n\x06status\x18\x05 \x01(\x0e\x32+.a3m.api.transferservice.v1beta1.Job.StatusR\x06status\x12\x39\n\nstart_time\x18\x06 \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tstartTime"_\n\x06Status\x12\x16\n\x12STATUS_UNSPECIFIED\x10\x00\x12\x13\n\x0fSTATUS_COMPLETE\x10\x01\x12\x15\n\x11STATUS_PROCESSING\x10\x02\x12\x11\n\rSTATUS_FAILED\x10\x03"\xc6\x02\n\x04Task\x12\x0e\n\x02id\x18\x01 \x01(\tR\x02id\x12\x17\n\x07\x66ile_id\x18\x02 \x01(\tR\x06\x66ileId\x12\x1b\n\texit_code\x18\x03 \x01(\x05R\x08\x65xitCode\x12\x1a\n\x08\x66ilename\x18\x04 \x01(\tR\x08\x66ilename\x12\x1c\n\texecution\x18\x05 \x01(\tR\texecution\x12\x1c\n\targuments\x18\x06 \x01(\tR\targuments\x12\x16\n\x06stdout\x18\x07 \x01(\tR\x06stdout\x12\x16\n\x06stderr\x18\x08 \x01(\tR\x06stderr\x12\x39\n\nstart_time\x18\t \x01(\x0b\x32\x1a.google.protobuf.TimestampR\tstartTime\x12\x35\n\x08\x65nd_time\x18\n \x01(\x0b\x32\x1a.google.protobuf.TimestampR\x07\x65ndTime"\xcc\n\n\x10ProcessingConfig\x12=\n\x1b\x61ssign_uuids_to_directories\x18\x01 \x01(\x08R\x18\x61ssignUuidsToDirectories\x12)\n\x10\x65xamine_contents\x18\x02 \x01(\x08R\x0f\x65xamineContents\x12K\n"generate_transfer_structure_report\x18\x03 \x01(\x08R\x1fgenerateTransferStructureReport\x12<\n\x1a\x64ocument_empty_directories\x18\x04 \x01(\x08R\x18\x64ocumentEmptyDirectories\x12)\n\x10\x65xtract_packages\x18\x05 \x01(\x08R\x0f\x65xtractPackages\x12G\n delete_packages_after_extraction\x18\x06 \x01(\x08R\x1d\x64\x65letePackagesAfterExtraction\x12+\n\x11identify_transfer\x18\x07 \x01(\x08R\x10identifyTransfer\x12G\n identify_submission_and_metadata\x18\x08 \x01(\x08R\x1didentifySubmissionAndMetadata\x12\x42\n\x1didentify_before_normalization\x18\t \x01(\x08R\x1bidentifyBeforeNormalization\x12\x1c\n\tnormalize\x18\n \x01(\x08R\tnormalize\x12)\n\x10transcribe_files\x18\x0b \x01(\x08R\x0ftranscribeFiles\x12J\n"perform_policy_checks_on_originals\x18\x0c \x01(\x08R\x1eperformPolicyChecksOnOriginals\x12g\n1perform_policy_checks_on_preservation_derivatives\x18\r \x01(\x08R,performPolicyChecksOnPreservationDerivatives\x12\x32\n\x15\x61ip_compression_level\x18\x0e \x01(\x05R\x13\x61ipCompressionLevel\x12\x85\x01\n\x19\x61ip_compression_algorithm\x18\x0f \x01(\x0e\x32I.a3m.api.transferservice.v1beta1.ProcessingConfig.AIPCompressionAlgorithmR\x17\x61ipCompressionAlgorithm"\xda\x02\n\x17\x41IPCompressionAlgorithm\x12)\n%AIP_COMPRESSION_ALGORITHM_UNSPECIFIED\x10\x00\x12*\n&AIP_COMPRESSION_ALGORITHM_UNCOMPRESSED\x10\x01\x12!\n\x1d\x41IP_COMPRESSION_ALGORITHM_TAR\x10\x02\x12\'\n#AIP_COMPRESSION_ALGORITHM_TAR_BZIP2\x10\x03\x12&\n"AIP_COMPRESSION_ALGORITHM_TAR_GZIP\x10\x04\x12%\n!AIP_COMPRESSION_ALGORITHM_S7_COPY\x10\x05\x12&\n"AIP_COMPRESSION_ALGORITHM_S7_BZIP2\x10\x06\x12%\n!AIP_COMPRESSION_ALGORITHM_S7_LZMA\x10\x07*\xa3\x01\n\rPackageStatus\x12\x1e\n\x1aPACKAGE_STATUS_UNSPECIFIED\x10\x00\x12\x19\n\x15PACKAGE_STATUS_FAILED\x10\x01\x12\x1b\n\x17PACKAGE_STATUS_REJECTED\x10\x02\x12\x1b\n\x17PACKAGE_STATUS_COMPLETE\x10\x03\x12\x1d\n\x19PACKAGE_STATUS_PROCESSING\x10\x04\x42\xb1\x02\n#com.a3m.api.transferservice.v1beta1B\x14RequestResponseProtoP\x01ZUgithub.com/artefactual-labs/a3m/proto/a3m/api/transferservice/v1beta1;transferservice\xa2\x02\x03\x41\x41T\xaa\x02\x1f\x41\x33m.Api.Transferservice.V1beta1\xca\x02\x1f\x41\x33m\\Api\\Transferservice\\V1beta1\xe2\x02+A3m\\Api\\Transferservice\\V1beta1\\GPBMetadata\xea\x02"A3m::Api::Transferservice::V1beta1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "a3m.api.transferservice.v1beta1.request_response_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\n#com.a3m.api.transferservice.v1beta1B\024RequestResponseProtoP\001ZUgithub.com/artefactual-labs/a3m/proto/a3m/api/transferservice/v1beta1;transferservice\242\002\003AAT\252\002\037A3m.Api.Transferservice.V1beta1\312\002\037A3m\\Api\\Transferservice\\V1beta1\342\002+A3m\\Api\\Transferservice\\V1beta1\\GPBMetadata\352\002"A3m::Api::Transferservice::V1beta1'
    _globals["_PACKAGESTATUS"]._serialized_start = 2648
    _globals["_PACKAGESTATUS"]._serialized_end = 2811
    _globals["_SUBMITREQUEST"]._serialized_start = 125
    _globals["_SUBMITREQUEST"]._serialized_end = 253
    _globals["_SUBMITRESPONSE"]._serialized_start = 255
    _globals["_SUBMITRESPONSE"]._serialized_end = 287
    _globals["_READREQUEST"]._serialized_start = 289
    _globals["_READREQUEST"]._serialized_end = 318
    _globals["_READRESPONSE"]._serialized_start = 321
    _globals["_READRESPONSE"]._serialized_end = 483
    _globals["_LISTTASKSREQUEST"]._serialized_start = 485
    _globals["_LISTTASKSREQUEST"]._serialized_end = 526
    _globals["_LISTTASKSRESPONSE"]._serialized_start = 528
    _globals["_LISTTASKSRESPONSE"]._serialized_end = 608
    _globals["_EMPTYREQUEST"]._serialized_start = 610
    _globals["_EMPTYREQUEST"]._serialized_end = 624
    _globals["_EMPTYRESPONSE"]._serialized_start = 626
    _globals["_EMPTYRESPONSE"]._serialized_end = 641
    _globals["_JOB"]._serialized_start = 644
    _globals["_JOB"]._serialized_end = 957
    _globals["_JOB_STATUS"]._serialized_start = 862
    _globals["_JOB_STATUS"]._serialized_end = 957
    _globals["_TASK"]._serialized_start = 960
    _globals["_TASK"]._serialized_end = 1286
    _globals["_PROCESSINGCONFIG"]._serialized_start = 1289
    _globals["_PROCESSINGCONFIG"]._serialized_end = 2645
    _globals["_PROCESSINGCONFIG_AIPCOMPRESSIONALGORITHM"]._serialized_start = 2299
    _globals["_PROCESSINGCONFIG_AIPCOMPRESSIONALGORITHM"]._serialized_end = 2645
# @@protoc_insertion_point(module_scope)



from pathlib import Path

from bagit import Bag
from bagit import BagError


def is_bag(path):
    """Determine whether the directory contains a BagIt package.

    The constructor of ``Bag`` is fast enough but we may prefer to optimize
    later.
    """
    if isinstance(path, Path):
        path = str(path)
    try:
        Bag(path)
    except BagError:
        return False
    return True



import os
import pathlib
import sys
import warnings

import django
from django.conf import settings


def suppress_warnings():
    """Suppress SyntaxWarning.

    Hiding SyntaxWarning from users since it can be misleading.
    """
    if settings.DEBUG or sys.warnoptions:
        return
    warnings.simplefilter("ignore", SyntaxWarning)


def init_django():
    """Initialize our Django project.

    Why do we need this? Django does not let us import models unless Django
    itself is set up. The alternative is lazy imports but we are not taking
    that approach at the moment.
    """
    os.environ.setdefault("DJANGO_SETTINGS_MODULE", "a3m.settings.common")
    django.setup()


def configure_xml_catalog_files():
    """Use local XML schemas for validation."""
    os.environ["XML_CATALOG_FILES"] = str(
        pathlib.Path(__file__).parent.parent / "client/assets/catalog/catalog.xml"
    )



"""
A Job is MCP Client's representation of a unit of work to be
performed--corresponding to a Task on the MCP Server side.  Jobs are run in
batches by clientScript modules and populated with an exit code, standard out
and standard error information.
"""
import logging
import os
import sys
import traceback
from contextlib import contextmanager


STANDARD_FORMAT = (
    "%(levelname)-8s  %(asctime)s  %(name)s.%(funcName)s:%(lineno)d  %(message)s"
)

SCRIPT_FILE_FORMAT = "{}: %(levelname)-8s  %(asctime)s  %(name)s:%(funcName)s:%(lineno)d:  %(message)s".format(
    os.path.basename(sys.argv[0])
)


class CallbackHandler(logging.Handler):
    def __init__(self, callback, module_name=None):
        logging.Handler.__init__(self)
        fmt = STANDARD_FORMAT if module_name else SCRIPT_FILE_FORMAT
        self.formatter = logging.Formatter(f"{module_name}: {fmt}")
        self.callback = callback

    def emit(self, record):
        self.callback(self.format(record))


class Job:
    def __init__(self, name, uuid, args, caller_wants_output=False):
        self.name = name
        self.UUID = uuid
        self.args = [name] + args
        self.caller_wants_output = caller_wants_output
        self.int_code = 0
        self.status_code = "success"
        self.output = ""
        self.error = ""

    def dump(self):
        return (
            "\n\n\t| =============== JOB\n"
            "\t| %s (exit=%s; code=%s uuid=%s)\n"
            "\t| =============== STDOUT\n"
            "\t| %s\n"
            "\t| =============== END STDOUT\n"
            "\t| =============== STDERR\n"
            "\t| %s\n"
            "\t| =============== END STDERR\n"
            "\t| =============== ARGS\n"
            "\t| %s\n"
            "\t| =============== END ARGS\n"
        ) % (
            self.name,
            self.int_code,
            self.status_code,
            self.UUID,
            self.get_stdout(),
            self.get_stderr(),
            self.args,
        )

    def set_status(self, int_code, status_code="success"):
        if int_code:
            self.int_code = int(int_code)
        self.status_code = status_code

    def write_output(self, s):
        self.output += s

    def write_error(self, s):
        self.error += s

    def print_output(self, *args):
        self.write_output(" ".join([self._to_str(x) for x in args]) + "\n")

    def print_error(self, *args):
        self.write_error(" ".join([self._to_str(x) for x in args]) + "\n")

    @staticmethod
    def _to_str(thing):
        try:
            return str(thing)
        except UnicodeEncodeError:
            return thing.encode("utf8")

    def pyprint(self, *objects, **kwargs):
        file = kwargs.get("file", sys.stdout)
        sep = kwargs.get("sep", " ")
        end = kwargs.get("end", "\n")
        msg = sep.join([self._to_str(x) for x in objects]) + end
        if file == sys.stdout:
            self.write_output(msg)
        elif file == sys.stderr:
            self.write_error(msg)
        else:
            raise Exception("Unrecognised print file: " + str(file))

    def get_exit_code(self):
        return self.int_code

    def get_stdout(self):
        return self.output

    def get_stderr(self):
        return self.error

    @contextmanager
    def JobContext(self, logger=None):
        handler = CallbackHandler(self.print_error, self.name)

        if logger:
            logger.addHandler(handler)

        try:
            yield
        except Exception as e:
            self.write_error(str(e))
            self.write_error(traceback.format_exc())
            self.set_status(1)
        finally:
            if logger:
                logger.removeHandler(handler)



"""
Format Policy Registry.

Rules serve as the central mechanisms that dictate how various file formats are
processed, each tailored for a specific format version and preservation purpose.
Each format version is an iteration of a broader format, categorized under a
general format group. The execution is carried out by commands, which are tied
to specific tools, ensuring every file is processed accurately. Commands are
adaptable, capable of being replaced and updated, and can even transform files
into different format versions as needed.

erDiagram

    Rule }o--o{ Rule : "replaces"
    Rule ||--o{ FormatVersion : "targets"
    Rule ||--o{ Command : "uses"

    FormatVersion }o--o{ FormatVersion: "replaces"
    FormatVersion ||--|| Format : "is a"

    Format ||--|| FormatGroup : "belongs to"

    Command }o--o{ Command: "replaces"
    Command }|--|| Tool : "uses"
    Command }o--o{ Command : "verifies"
    Command }o--o{ Command : "event details"
    Command ||--o{ FormatVersion : "outputs to"

Aspirationally, we would like to introduce the concept of a profile, i.e. a set
of rules and related objects offering a customzied approach to the processing of
digital files.
"""
from __future__ import annotations

import json
import uuid
from abc import ABC
from abc import abstractmethod
from datetime import datetime
from enum import Enum
from importlib.resources import files
from typing import Any
from typing import Literal
from typing import Protocol
from typing import TYPE_CHECKING
from typing import TypedDict
from typing import TypeVar

from django.apps import apps


# Avoid issues with circular imports.
if TYPE_CHECKING:
    from a3m.main.models import File
else:
    File = "File"


class RulePurpose(Enum):
    ACCESS = "access"
    CHARACTERIZATION = "characterization"
    CHARACTERIZE = "characterize"
    DEFAULT_ACCESS = "default_access"
    DEFAULT_CHARACTERIZATION = "default_characterization"
    DEFAULT_THUMBNAIL = "default_thumbnail"
    EXTRACT = "extract"
    POLICY_CHECK = "policy_check"
    PRESERVATION = "preservation"
    THUMBNAIL = "thumbnail"
    TRANSCRIPTION = "transcription"
    VALIDATION = "validation"

    def get_fallback(self) -> RulePurpose | None:
        if self is RulePurpose.THUMBNAIL:
            return RulePurpose.DEFAULT_THUMBNAIL
        if self is RulePurpose.CHARACTERIZATION:
            return RulePurpose.DEFAULT_CHARACTERIZATION
        if self is RulePurpose.ACCESS:
            return RulePurpose.DEFAULT_ACCESS
        return None


class CommandScriptType(Enum):
    AS_IS = "as_is"
    BASH_SCRIPT = "bashScript"
    COMMAND = "command"
    PYTHON_SCRIPT = "pythonScript"


class CommandUsage(Enum):
    CHARACTERIZATION = "characterization"
    EVENT_DETAIL = "event_detail"
    EXTRACTION = "extraction"
    NORMALIZATION = "normalization"
    TRANSCRIPTION = "transcription"
    VALIDATION = "validation"
    VERIFICATION = "verification"


class FormatGroup:
    """
    "fields": {
        "slug": "audio",
        "uuid": "c94ce0e6-c275-4c09-b802-695a18b7bf2a",
        "description": "Audio"
    }
    """

    id: uuid.UUID
    slug: str
    description: str


class Format:
    """
    "fields": {
        "slug": "dolby-digital-audio",
        "group": "c94ce0e6-c275-4c09-b802-695a18b7bf2a",
        "uuid": "9426449b-6ead-4b90-a3f9-0e2f21f3bf8b",
        "description": "Dolby Digital Audio"
    }
    """

    id: uuid.UUID
    group: FormatGroup
    slug: str
    description: str


class FormatVersion:
    """
    "fields": {
        "replaces": null,
        "uuid": "8943dfdc-2b37-47df-b00f-76472e92e56d",
        "format": "ef618927-7376-4088-a761-182f7ab796bf",
        "lastmodified": "2013-11-15T01:18:30Z",
        "enabled": true,
        "access_format": false,
        "preservation_format": false,
        "version": "",
        "slug": "generic-gif",
        "pronom_id": "",
        "description": "Generic gif"
    }
    """

    id: uuid.UUID
    replaces: uuid.UUID | None
    format: Format
    last_modified: datetime
    enabled: bool
    access_format: bool
    preservation_format: bool
    version: str
    slug: str
    pronom_id: str
    description: str


class Tool:
    """
    "fields": {
        "slug": "inkscape-04831-r9886-jan-29-2013",
        "version": "0.48.3.1 r9886 (Jan 29 2013)",
        "enabled": false,
        "uuid": "06523365-9061-4ac7-a7f5-3e7e5f06ffac",
        "description": "inkscape"
    },
    """

    id: uuid.UUID
    slug: str
    version: str
    enabled: bool
    description: str


class Command:
    """
    "fields": {
        "replaces": null,
        "uuid": "6537147f-4dd4-4950-8aff-5578db9a485d",
        "lastmodified": "2013-11-15T01:18:36Z",
        "tool": "c5465b07-8dc7-475e-a5c9-ccb2ba2ed083",
        "enabled": false,
        "event_detail_command": null,
        "output_location": "",
        "command_usage": "characterization",
        "verification_command": null,
        "command": "This is a placeholder command only, and should not be called.",
        "script_type": "as_is",
        "output_format": "d60e5243-692e-4af7-90cd-40c53cb8dc7d",
        "description": "FITS"
    }
    """

    id: uuid.UUID
    replaces: uuid.UUID | None
    last_modified: datetime
    enabled: bool
    output_location: str
    command_usage: CommandUsage
    command: str
    script_type: CommandScriptType
    description: str
    tool: Tool
    output_format: FormatVersion | None
    verification_command: Command | None
    event_detail_command: Command | None


class Rule:
    """
    "fields": {
        "replaces": null,
        "uuid": "50ec04f1-372d-435d-a7c6-a548e7006d74",
        "format": "0ab4cd40-90e7-4d75-b294-498177b3897d",
        "lastmodified": "2013-11-15T01:18:38Z",
        "enabled": true,
        "command": "41112047-7ddf-4bf0-9156-39fe96b32d53",
        "purpose": "default_access"
    }
    """

    id: uuid.UUID
    replaces: uuid.UUID | None
    purpose: RulePurpose
    command: Command
    format: FormatVersion
    last_modified: datetime
    enabled: bool


class Backend(ABC):
    """Represents the contract of FPR backends.

    Backends must omit disabled or replaced objects.
    """

    @abstractmethod
    def get_format_version_by_id(self, id: uuid.UUID) -> FormatVersion | None:
        pass

    @abstractmethod
    def get_format_version_by_puid(self, puid: str) -> FormatVersion | None:
        pass

    @abstractmethod
    def get_rule_by_id(self, id: uuid.UUID) -> Rule | None:
        pass

    @abstractmethod
    def get_rules(
        self, format_version_id: uuid.UUID, purpose: RulePurpose
    ) -> list[Rule]:
        pass


class Replaceable(Protocol):
    id: uuid.UUID
    replaces: uuid.UUID | None


T = TypeVar("T", bound=Replaceable)


def get_replaced_objects(objects: dict[uuid.UUID, T]) -> dict[uuid.UUID, uuid.UUID]:
    return {obj.replaces: obj.id for obj in objects.values() if obj.replaces}


class DjangoDatabaseDumpItem(TypedDict):
    """Represents database items in a Django sqldump."""

    fields: dict[str, Any]
    model: Literal[
        "fpr.format",
        "fpr.formatgroup",
        "fpr.formatversion",
        "fpr.fpcommand",
        "fpr.fprule",
        "fpr.fptool",
    ]
    pk: int


DjangoDatabaseDumpList = list[DjangoDatabaseDumpItem]


class JSONBackend(Backend):
    formats: dict[uuid.UUID, Format]
    groups: dict[uuid.UUID, FormatGroup]
    versions: dict[uuid.UUID, FormatVersion]
    commands: dict[uuid.UUID, Command]
    rules: dict[uuid.UUID, Rule]
    tools: dict[uuid.UUID, Tool]

    replaced_versions: dict[uuid.UUID, uuid.UUID]
    replaced_commands: dict[uuid.UUID, uuid.UUID]
    replaced_rules: dict[uuid.UUID, uuid.UUID]

    def __init__(self, blob):
        self._load(json.loads(blob))

    def _load(self, data: DjangoDatabaseDumpList):
        self.formats = {}
        self.groups = {}
        self.versions = {}
        self.commands = {}
        self.rules = {}
        self.tools = {}

        # Load format groups.
        for item in data:
            model, fields = item["model"], item["fields"]
            id = uuid.UUID(fields["uuid"])
            if model == "fpr.formatgroup":
                group = FormatGroup()
                self.groups[id] = group
                group.id = id
                group.slug = fields["slug"]
                group.description = fields["description"]

        # Load formats.
        for item in data:
            model, fields = item["model"], item["fields"]
            id = uuid.UUID(fields["uuid"])
            if model == "fpr.format":
                format = Format()
                self.formats[id] = format
                format.id = id
                format.slug = fields["slug"]
                format.description = fields["description"]
                format.group = self.groups[uuid.UUID(fields["group"])]

        # Load format versions.
        for item in data:
            model, fields = item["model"], item["fields"]
            id = uuid.UUID(fields["uuid"])
            if model == "fpr.formatversion":
                version = FormatVersion()
                self.versions[id] = version
                version.id = id
                version.replaces = (
                    uuid.UUID(fields["replaces"]) if fields["replaces"] else None
                )
                version.last_modified = convert_django_date_to_datetime(
                    fields["lastmodified"]
                )
                version.enabled = fields["enabled"]
                version.access_format = fields["access_format"]
                version.preservation_format = fields["preservation_format"]
                version.version = fields["version"]
                version.slug = fields["slug"]
                version.pronom_id = fields["pronom_id"]
                version.description = fields["description"]
                version.format = self.formats[uuid.UUID(fields["format"])]
        self.replaced_versions = get_replaced_objects(self.versions)

        # Load tools.
        for item in data:
            model, fields = item["model"], item["fields"]
            id = uuid.UUID(fields["uuid"])
            if model == "fpr.fptool":
                tool = Tool()
                self.tools[id] = tool
                tool.id = id
                tool.slug = fields["slug"]
                tool.description = fields["description"]
                tool.version = fields["version"]

        # Load commands.
        for item in data:
            model, fields = item["model"], item["fields"]
            id = uuid.UUID(fields["uuid"])
            if model == "fpr.fpcommand":
                command = Command()
                self.commands[id] = command
                command.id = id
                command.replaces = (
                    uuid.UUID(fields["replaces"]) if fields["replaces"] else None
                )
                command.enabled = fields["enabled"]
                command.last_modified = convert_django_date_to_datetime(
                    fields["lastmodified"]
                )
                command.output_location = fields["output_location"]
                command.command = fields["command"]
                command.command_usage = CommandUsage(fields["command_usage"].casefold())
                command.script_type = CommandScriptType(fields["script_type"])
                command.description = fields["description"]
                command.tool = self.tools[uuid.UUID(fields["tool"])]
                if fields["output_format"]:
                    format_version_id = uuid.UUID(fields["output_format"])
                    command.output_format = self.versions.get(format_version_id)
        self.replaced_commands = get_replaced_objects(self.commands)
        for item in data:
            model, fields = item["model"], item["fields"]
            id = uuid.UUID(fields["uuid"])
            if model == "fpr.fpcommand":
                command = self.commands[id]
                command.verification_command = self.commands.get(
                    fields["verification_command"]
                )
                command.event_detail_command = self.commands.get(
                    fields["event_detail_command"]
                )

        # Load rules.
        for item in data:
            model, fields = item["model"], item["fields"]
            id = uuid.UUID(fields["uuid"])
            if model == "fpr.fprule":
                rule = Rule()
                rule.id = id
                rule.last_modified = convert_django_date_to_datetime(
                    fields["lastmodified"]
                )
                rule.replaces = (
                    uuid.UUID(fields["replaces"]) if fields["replaces"] else None
                )
                rule.enabled = fields["enabled"]
                rule.purpose = RulePurpose(fields["purpose"])
                try:
                    rule.format = self.versions[uuid.UUID(fields["format"])]
                    rule.command = self.commands[uuid.UUID(fields["command"])]
                except KeyError:
                    pass
                else:
                    self.rules[id] = rule
        self.replaced_rules = get_replaced_objects(self.rules)

    def get_format_version_by_id(self, id: uuid.UUID) -> FormatVersion | None:
        version = self.versions.get(id)
        if (
            version is None
            or not version.enabled
            or version.id in self.replaced_versions
        ):
            return None
        return version

    def get_format_version_by_puid(self, puid: str) -> FormatVersion | None:
        for version in self.versions.values():
            if (
                version.pronom_id == puid
                and version.enabled
                and version.id not in self.replaced_versions
            ):
                return version
        return None

    def get_rule_by_id(self, id: uuid.UUID) -> Rule | None:
        return self.rules.get(id)

    def get_rules(
        self, format_version_id: uuid.UUID, purpose: RulePurpose
    ) -> list[Rule]:
        rules: list[Rule] = []
        format_version = self.versions.get(format_version_id)
        if not format_version:
            raise ValueError(f"Format version f{format_version_id} not found.")
        for rule in self.rules.values():
            if rule.format.id == format_version_id:
                if rule.purpose == purpose:
                    if rule.enabled and rule.id not in self.replaced_rules:
                        rules.append(rule)
        return rules


def convert_django_date_to_datetime(date: str) -> datetime:
    """Convert a Django date string to a datetime object."""
    for fmt in ("%Y-%m-%dT%H:%M:%SZ", "%Y-%m-%dT%H:%M:%S.%fZ"):
        try:
            return datetime.strptime(date, fmt)
        except ValueError:
            continue
    raise ValueError(f"String '{date}' does not match any of the expected formats.")


def convert_to_uuid(id: uuid.UUID | str) -> uuid.UUID:
    """Ensure that a string is a UUID object."""
    if isinstance(id, str):
        return uuid.UUID(id)
    return id


class Registry:
    """FPR registry with pluggable data sources (backends).

    It omits objects not in service (disabled or replaced).

    It does not import Django models to avoid circular dependency issues.
    """

    backend: Backend

    def __init__(self, backend: Backend | None = None):
        self.backend = backend or default_backend

    def _file_model(self) -> File:
        return apps.get_model("main.File")  # type: ignore

    def get_format_version_by_id(self, id: uuid.UUID) -> FormatVersion | None:
        """Returns a format version given its identifier.

        It returns None when the object is not found or is not in service.
        """
        return self.backend.get_format_version_by_id(id)

    def get_format_version_by_puid(self, puid: str) -> FormatVersion | None:
        """Returns a format version given its PRONOM identifier.

        It returns None when the object is not found or is not in service.
        """
        return self.backend.get_format_version_by_puid(puid)

    def get_rule_by_id(self, id: uuid.UUID | str) -> Rule | None:
        """Returns a rule given its identifier.

        It returns None when the object is not found or is not in service.
        """
        return self.backend.get_rule_by_id(convert_to_uuid(id))

    def get_rules(
        self, format_version_id: uuid.UUID | str, purpose: RulePurpose
    ) -> list[Rule]:
        """Return the rules for a given format version and rule purpose.

        It omits rules not in service."""
        return self.backend.get_rules(convert_to_uuid(format_version_id), purpose)

    def get_file_rules(
        self, file: uuid.UUID | str | File, purpose: RulePurpose | str, fallback=False
    ) -> list[Rule]:
        """Return the rules for a given file and rule purpose.

        It omits rules not in service.

        TODO: refactor as method of the File mode, should also typing and circular import issues.
        """
        result: list[Rule] = []
        if isinstance(purpose, str):
            purpose = RulePurpose(purpose)
        file_obj: File
        if str(type(file).__name__) == File:
            file_obj = file  # type: ignore
        elif isinstance(file, uuid.UUID):
            file_obj = self._file_model().objects.get(pk=str(file))
        elif isinstance(file, str):
            file_obj = self._file_model().objects.get(pk=file)
        else:
            raise ValueError
        for format_version_id in file_obj.fileformatversion_set.values_list(
            "format_version_id", flat=True
        ):
            rules = self.get_rules(format_version_id, purpose)
            if len(rules):
                result.extend(rules)
            elif fallback:
                if (fallback_purpose := purpose.get_fallback()) is not None:
                    result.extend(
                        self.get_rules(format_version_id, purpose=fallback_purpose)
                    )
        return result


default_backend = JSONBackend(
    files("a3m.fpr.migrations").joinpath("initial-data.json").read_bytes()
)


FPR = Registry(default_backend)



"""Shared directory setup.
"""
import logging
import os
import shutil

from django.conf import settings


logger = logging.getLogger(__name__)


def create():
    dirs = (
        "currentlyProcessing/transfer",
        "currentlyProcessing/ingest",
        "completed",
        "failed",
        "policies",
        "tmp",
    )
    for dirname in dirs:
        dirname = os.path.join(settings.SHARED_DIRECTORY, dirname)
        if os.path.isdir(dirname):
            continue
        logger.debug("Creating directory: %s", dirname)
        os.makedirs(dirname, mode=0o770)


def empty():
    dirs = (
        "currentlyProcessing",
        "completed",
        "failed",
        "policies",
        "tmp",
    )
    for dirname in dirs:
        dirname = os.path.join(settings.SHARED_DIRECTORY, dirname)
        if os.path.isdir(dirname):
            logger.debug("Removing directory and contents: %s", dirname)
            shutil.rmtree(dirname, ignore_errors=True)



"""
Jobs relating to configurable decisions.
"""
import logging

from a3m.server.jobs.base import Job


logger = logging.getLogger(__name__)


class NextLinkDecisionJob(Job):
    """A job that determines the next link to be executed."""

    def run(self, *args, **kwargs):
        super().run(*args, **kwargs)

        logger.debug("Running %s (package %s)", self.description, self.package.uuid)

        # Reload the package, in case the path has changed
        self.package.reload()
        self.save_to_db()

        self.job_chain.next_link = self.decide()

        return next(self.job_chain, None)

    def decide(self):
        config = self.link.config

        config_value = self.get_configured_value(config["config_attr"])
        if config_value is None:
            config_value = config["default"]

        next_id = None
        for item in config["choices"]:
            if item["value"] == config_value:
                next_id = item["link_id"]
                break

        logger.debug("Using user selected link %s", next_id)
        self.mark_complete()

        return next_id

    def get_configured_value(self, attr_name):
        return getattr(self.package.config, attr_name, None)



import math
import time
from contextlib import contextmanager

from prometheus_client import Counter

# We need to balance reasonably accurate tracking with high cardinality here, as
# this is used with script_name labels and there are already over 100 scripts.
TASK_DURATION_BUCKETS = (
    2.0,
    5.0,
    10.0,
    20.0,
    30.0,
    60.0,
    120.0,  # 2 min
    300.0,  # 5 min
    600.0,  # 10 min
    1800.0,  # 30 min
    3600.0,  # 1 hour
    math.inf,
)
# Histogram buckets for total processing time, e.g. for an AIP.
# Not used with labels.
PROCESSING_TIME_BUCKETS = (
    10.0,
    20.0,
    30.0,
    60.0,
    120.0,  # 2 min
    300.0,  # 5 min
    600.0,  # 10 min
    1800.0,  # 30 min
    3600.0,  # 1 hour
    7200.0,  # 2 hours
    14400.0,  # 4 hours
    28800.0,  # 8 hours
    math.inf,
)
# Histogram for distribution of transfer and AIP file counts
PACKAGE_FILE_COUNT_BUCKETS = (
    10.0,
    50.0,
    100.0,
    250.0,
    500.0,
    1000.0,
    2000.0,
    5000.0,
    10000.0,
    math.inf,
)
# Histogram for distribution of transfer and AIP size in bytes
PACKAGE_SIZE_BUCKETS = (
    1000000.0,  # 1 MB
    10000000.0,  # 10 MB
    50000000.0,  # 50 MB
    100000000.0,  # 100 MB
    200000000.0,  # 200 MB
    500000000.0,  # 500 MB
    1000000000.0,  # 1 GB
    5000000000.0,  # 5 GB
    10000000000.0,  # 10 GB
    math.inf,
)


db_retry_time_counter = Counter(
    "common_db_retry_time_seconds",
    (
        "Total time waiting to retry database transactions in seconds, labeled "
        "by operation description"
    ),
    ["description"],
)


@contextmanager
def db_retry_timer(*args, **kwargs):
    start_time = time.time()
    try:
        yield
    finally:
        duration = time.time() - start_time
        db_retry_time_counter.labels(**kwargs).inc(duration)



#
# This file is part of Archivematica.
#
# Copyright 2010-2013 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.    See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.    If not, see <http://www.gnu.org/licenses/>.
import sys
import uuid

import lxml.etree as etree

from a3m import namespaces as ns
from a3m.countryCodes import getCodeForCountry
from a3m.main.models import RightsStatement

RIGHTS_BASIS_OTHER = ["Policy", "Donor"]


def formatDate(date):
    """hack fix for 0.8, easy dashboard insertion ISO 8601 -> edtfSimpleType"""
    if date:
        date = date.replace("/", "-")
    return date


def _add_start_end_date_complex_type(name, elem, start_date, end_date, end_date_open):
    """Add new ``startAndEndDateComplexType`` subelement to ``elem``.

    The following elements in PREMIS use it:

      - ``copyrightApplicableDates``
      - ``licenseApplicableDates``
      - ``otherRightsApplicableDates``
      - ``statuteApplicableDates``
      - ``termOfGrant``
      - ``termOfRestriction``

    """
    if not any([start_date, end_date, end_date_open]):
        return
    dates = etree.SubElement(elem, ns.premisBNS + name)
    etree.SubElement(dates, ns.premisBNS + "startDate").text = (
        formatDate(start_date) if start_date else ""
    )
    end_date = "OPEN" if end_date_open else formatDate(end_date)
    if end_date:
        etree.SubElement(dates, ns.premisBNS + "endDate").text = end_date


def archivematicaGetRights(job, metadataAppliesToList, fileUUID, state):
    """[(fileUUID, fileUUIDTYPE), (sipUUID, sipUUIDTYPE), (transferUUID, transferUUIDType)]"""
    ret = []
    for metadataAppliesToidentifier, metadataAppliesToType in metadataAppliesToList:
        statements = RightsStatement.objects.filter(
            metadataappliestoidentifier=metadataAppliesToidentifier,
            metadataappliestotype_id=metadataAppliesToType,
        )
        for statement in statements:
            rightsStatement = createRightsStatement(job, statement, fileUUID, state)
            ret.append(rightsStatement)
    return ret


def createRightsStatement(job, statement, fileUUID, state):
    rightsStatement = etree.Element(
        ns.premisBNS + "rightsStatement", nsmap={"premis": ns.premisNS}
    )
    rightsStatement.set(
        ns.xsiBNS + "schemaLocation",
        ns.premisNS + " http://www.loc.gov/standards/premis/v3/premis.xsd",
    )

    # rightsStatement.set("version", "2.1") # cvc-complex-type.3.2.2: Attribute 'version' is not allowed to appear in element 'rightsStatement'.

    rightsStatementIdentifier = etree.SubElement(
        rightsStatement, ns.premisBNS + "rightsStatementIdentifier"
    )
    if statement.rightsstatementidentifiervalue:
        etree.SubElement(
            rightsStatementIdentifier, ns.premisBNS + "rightsStatementIdentifierType"
        ).text = statement.rightsstatementidentifiertype
        etree.SubElement(
            rightsStatementIdentifier, ns.premisBNS + "rightsStatementIdentifierValue"
        ).text = statement.rightsstatementidentifiervalue
    else:
        etree.SubElement(
            rightsStatementIdentifier, ns.premisBNS + "rightsStatementIdentifierType"
        ).text = "UUID"
        etree.SubElement(
            rightsStatementIdentifier, ns.premisBNS + "rightsStatementIdentifierValue"
        ).text = uuid.uuid4().__str__()
    if statement.rightsbasis in RIGHTS_BASIS_OTHER:
        etree.SubElement(rightsStatement, ns.premisBNS + "rightsBasis").text = "Other"
    else:
        etree.SubElement(
            rightsStatement, ns.premisBNS + "rightsBasis"
        ).text = statement.rightsbasis

    # Copyright information
    if statement.rightsbasis.lower() in ["copyright"]:
        for copyright in statement.rightsstatementcopyright_set.all():
            copyrightInformation = etree.SubElement(
                rightsStatement, ns.premisBNS + "copyrightInformation"
            )
            etree.SubElement(
                copyrightInformation, ns.premisBNS + "copyrightStatus"
            ).text = copyright.copyrightstatus
            copyrightJurisdiction = copyright.copyrightjurisdiction
            copyrightJurisdictionCode = getCodeForCountry(
                copyrightJurisdiction.__str__().upper()
            )
            if copyrightJurisdictionCode is not None:
                copyrightJurisdiction = copyrightJurisdictionCode
            etree.SubElement(
                copyrightInformation, ns.premisBNS + "copyrightJurisdiction"
            ).text = copyrightJurisdiction
            etree.SubElement(
                copyrightInformation, ns.premisBNS + "copyrightStatusDeterminationDate"
            ).text = formatDate(copyright.copyrightstatusdeterminationdate)
            # copyrightNote Repeatable
            for note in copyright.rightsstatementcopyrightnote_set.all():
                etree.SubElement(
                    copyrightInformation, ns.premisBNS + "copyrightNote"
                ).text = note.copyrightnote

            # RightsStatementCopyrightDocumentationIdentifier
            getDocumentationIdentifier(copyright, copyrightInformation)

            _add_start_end_date_complex_type(
                "copyrightApplicableDates",
                copyrightInformation,
                copyright.copyrightapplicablestartdate,
                copyright.copyrightapplicableenddate,
                copyright.copyrightenddateopen,
            )

    elif statement.rightsbasis.lower() in ["license"]:
        for license in statement.rightsstatementlicense_set.all():
            licenseInformation = etree.SubElement(
                rightsStatement, ns.premisBNS + "licenseInformation"
            )

            for (
                identifier
            ) in license.rightsstatementlicensedocumentationidentifier_set.all():
                licenseDocumentIdentifier = etree.SubElement(
                    licenseInformation, ns.premisBNS + "licenseDocumentationIdentifier"
                )
                etree.SubElement(
                    licenseDocumentIdentifier,
                    ns.premisBNS + "licenseDocumentationIdentifierType",
                ).text = identifier.licensedocumentationidentifiertype
                etree.SubElement(
                    licenseDocumentIdentifier,
                    ns.premisBNS + "licenseDocumentationIdentifierValue",
                ).text = identifier.licensedocumentationidentifiervalue
                etree.SubElement(
                    licenseDocumentIdentifier, ns.premisBNS + "licenseDocumentationRole"
                ).text = identifier.licensedocumentationidentifierrole

            etree.SubElement(
                licenseInformation, ns.premisBNS + "licenseTerms"
            ).text = license.licenseterms

            for note in license.rightsstatementlicensenote_set.all():
                etree.SubElement(
                    licenseInformation, ns.premisBNS + "licenseNote"
                ).text = note.licensenote

            _add_start_end_date_complex_type(
                "licenseApplicableDates",
                licenseInformation,
                license.licenseapplicablestartdate,
                license.licenseapplicableenddate,
                license.licenseenddateopen,
            )

    elif statement.rightsbasis.lower() in ["statute"]:
        # 4.1.5 statuteInformation (O, R)
        getstatuteInformation(statement, rightsStatement)

    elif statement.rightsbasis.lower() in ["donor", "policy", "other"]:
        otherRightsInformation = etree.SubElement(
            rightsStatement, ns.premisBNS + "otherRightsInformation"
        )
        for info in statement.rightsstatementotherrightsinformation_set.all():
            # otherRightsDocumentationIdentifier
            for (
                identifier
            ) in info.rightsstatementotherrightsdocumentationidentifier_set.all():
                otherRightsDocumentationIdentifier = etree.SubElement(
                    otherRightsInformation,
                    ns.premisBNS + "otherRightsDocumentationIdentifier",
                )
                etree.SubElement(
                    otherRightsDocumentationIdentifier,
                    ns.premisBNS + "otherRightsDocumentationIdentifierType",
                ).text = identifier.otherrightsdocumentationidentifiertype
                etree.SubElement(
                    otherRightsDocumentationIdentifier,
                    ns.premisBNS + "otherRightsDocumentationIdentifierValue",
                ).text = identifier.otherrightsdocumentationidentifiervalue
                etree.SubElement(
                    otherRightsDocumentationIdentifier,
                    ns.premisBNS + "otherRightsDocumentationRole",
                ).text = identifier.otherrightsdocumentationidentifierrole

            otherRightsBasis = info.otherrightsbasis

            if (
                not otherRightsBasis or statement.rightsbasis in RIGHTS_BASIS_OTHER
            ):  # not 100%
                otherRightsBasis = statement.rightsbasis
            etree.SubElement(
                otherRightsInformation, ns.premisBNS + "otherRightsBasis"
            ).text = otherRightsBasis

            _add_start_end_date_complex_type(
                "otherRightsApplicableDates",
                otherRightsInformation,
                info.otherrightsapplicablestartdate,
                info.otherrightsapplicableenddate,
                info.otherrightsenddateopen,
            )

            # otherRightsNote Repeatable
            for note in info.rightsstatementotherrightsinformationnote_set.all():
                etree.SubElement(
                    otherRightsInformation, ns.premisBNS + "otherRightsNote"
                ).text = note.otherrightsnote

    # 4.1.6 rightsGranted (O, R)
    getrightsGranted(job, statement, rightsStatement, state)

    # 4.1.7 linkingObjectIdentifier (O, R)
    linkingObjectIdentifier = etree.SubElement(
        rightsStatement, ns.premisBNS + "linkingObjectIdentifier"
    )
    etree.SubElement(
        linkingObjectIdentifier, ns.premisBNS + "linkingObjectIdentifierType"
    ).text = "UUID"
    etree.SubElement(
        linkingObjectIdentifier, ns.premisBNS + "linkingObjectIdentifierValue"
    ).text = fileUUID
    return rightsStatement


def getDocumentationIdentifier(copyright, parent):
    for (
        identifier
    ) in copyright.rightsstatementcopyrightdocumentationidentifier_set.all():
        statuteInformation = etree.SubElement(
            parent, ns.premisBNS + "copyrightDocumentationIdentifier"
        )
        etree.SubElement(
            statuteInformation, ns.premisBNS + "copyrightDocumentationIdentifierType"
        ).text = identifier.copyrightdocumentationidentifiertype
        etree.SubElement(
            statuteInformation, ns.premisBNS + "copyrightDocumentationIdentifierValue"
        ).text = identifier.copyrightdocumentationidentifiervalue
        etree.SubElement(
            statuteInformation, ns.premisBNS + "copyrightDocumentationRole"
        ).text = identifier.copyrightdocumentationidentifierrole


def getstatuteInformation(statement, parent):
    for statute in statement.rightsstatementstatuteinformation_set.all():
        statuteInformation = etree.SubElement(
            parent, ns.premisBNS + "statuteInformation"
        )
        etree.SubElement(
            statuteInformation, ns.premisBNS + "statuteJurisdiction"
        ).text = statute.statutejurisdiction
        etree.SubElement(
            statuteInformation, ns.premisBNS + "statuteCitation"
        ).text = statute.statutecitation
        etree.SubElement(
            statuteInformation, ns.premisBNS + "statuteInformationDeterminationDate"
        ).text = formatDate(statute.statutedeterminationdate)

        # statuteNote Repeatable
        for note in statute.rightsstatementstatuteinformationnote_set.all():
            etree.SubElement(
                statuteInformation, ns.premisBNS + "statuteNote"
            ).text = note.statutenote

        for (
            identifier
        ) in statute.rightsstatementstatutedocumentationidentifier_set.all():
            statuteDocumentationIdentifier = etree.SubElement(
                statuteInformation, ns.premisBNS + "statuteDocumentationIdentifier"
            )
            etree.SubElement(
                statuteDocumentationIdentifier,
                ns.premisBNS + "statuteDocumentationIdentifierType",
            ).text = identifier.statutedocumentationidentifiertype
            etree.SubElement(
                statuteDocumentationIdentifier,
                ns.premisBNS + "statuteDocumentationIdentifierValue",
            ).text = identifier.statutedocumentationidentifiervalue
            etree.SubElement(
                statuteDocumentationIdentifier,
                ns.premisBNS + "statuteDocumentationRole",
            ).text = identifier.statutedocumentationidentifierrole

        _add_start_end_date_complex_type(
            "statuteApplicableDates",
            statuteInformation,
            statute.statuteapplicablestartdate,
            statute.statuteapplicableenddate,
            statute.statuteenddateopen,
        )


def getrightsGranted(job, statement, parent, state):
    for granted in statement.rightsstatementrightsgranted_set.all():
        rightsGranted = etree.SubElement(parent, ns.premisBNS + "rightsGranted")
        etree.SubElement(rightsGranted, ns.premisBNS + "act").text = granted.act

        restriction = "Undefined"
        for restriction in granted.restrictions.all():
            restriction = restriction.restriction
            if restriction.lower() not in ["disallow", "conditional", "allow"]:
                job.pyprint(
                    "The value of element restriction must be: 'Allow', 'Disallow', or 'Conditional':",
                    restriction,
                    file=sys.stderr,
                )
                state.error_accumulator.error_count += 1
            etree.SubElement(
                rightsGranted, ns.premisBNS + "restriction"
            ).text = restriction

        if granted.startdate or granted.enddate or granted.enddateopen:
            restriction = restriction.lower()
            if restriction in ("allow",):
                term = "termOfGrant"
            elif restriction in ("disallow", "conditional"):
                term = "termOfRestriction"
            else:
                job.pyprint(
                    "The value of element restriction must be: 'Allow', "
                    "'Disallow', or 'Conditional'",
                    file=sys.stderr,
                )
                state.error_accumulator.error_count += 1
                continue
            _add_start_end_date_complex_type(
                term,
                rightsGranted,
                granted.startdate,
                granted.enddate,
                granted.enddateopen,
            )

        # 4.1.6.4 rightsGrantedNote (O, R)
        for note in granted.notes.all():
            etree.SubElement(
                rightsGranted, ns.premisBNS + "rightsGrantedNote"
            ).text = note.rightsgrantednote



from a3m.api.transferservice import v1beta1 as transfer_service_api

DEFAULT_PROCESSING_CONFIG = transfer_service_api.request_response_pb2.ProcessingConfig(
    assign_uuids_to_directories=True,
    examine_contents=False,
    generate_transfer_structure_report=True,
    document_empty_directories=True,
    extract_packages=True,
    delete_packages_after_extraction=False,
    identify_transfer=True,
    identify_submission_and_metadata=True,
    identify_before_normalization=True,
    normalize=True,
    transcribe_files=True,
    perform_policy_checks_on_originals=True,
    perform_policy_checks_on_preservation_derivatives=True,
    aip_compression_level=1,
    aip_compression_algorithm=transfer_service_api.request_response_pb2.ProcessingConfig.AIP_COMPRESSION_ALGORITHM_S7_COPY,
)



import argparse
import os.path
import sys

from django.db import transaction

from a3m import databaseFunctions
from a3m.api.transferservice.v1beta1.request_response_pb2 import ProcessingConfig
from a3m.executeOrRunSubProcess import executeOrRun
from a3m.main.models import SIP


def update_unit(sip_uuid, compressed_location):
    # Set aipFilename in Unit
    SIP.objects.filter(uuid=sip_uuid).update(
        aip_filename=os.path.basename(compressed_location)
    )


def compress_aip(
    job, compression, compression_level, sip_directory, sip_name, sip_uuid
):
    """Compresses AIP according to compression algorithm and level.
    compression = AIP compression algorithm, format: <program>-<algorithm>, eg. 7z-lzma, pbzip2-
    compression_level = AIP compression level, integer between 1 and 9 inclusive
    sip_directory = Absolute path to the directory where the SIP is
    sip_name = User-provided name of the SIP
    sip_uuid = SIP UUID

    Example inputs:
    compressAIP.py
        7z-lzma
        5
        %sharedDirectory%/watchedDirectories/workFlowDecisions/compressionAIPDecisions/ep-d87d5845-bd07-4200-b1a4-928e0cb6e1e4/
        ep
        d87d5845-bd07-4200-b1a4-928e0cb6e1e4
    """
    if compression_level == "0":
        compression_level = "1"

    # Default is uncompressed.
    compression = int(compression)
    ProcessingConfig.AIPCompressionAlgorithm.Name(compression)
    if compression == ProcessingConfig.AIP_COMPRESSION_ALGORITHM_UNSPECIFIED:
        compression = ProcessingConfig.AIP_COMPRESSION_ALGORITHM_UNCOMPRESSED

    # Translation to make compress_aip happy.
    mapping = {
        ProcessingConfig.AIP_COMPRESSION_ALGORITHM_UNCOMPRESSED: ("None", ""),
        ProcessingConfig.AIP_COMPRESSION_ALGORITHM_TAR: (
            "gzip",
            "tar.gzip",
        ),  # A3M-TODO: support
        ProcessingConfig.AIP_COMPRESSION_ALGORITHM_TAR_BZIP2: ("pbzip2", "pbzip2"),
        ProcessingConfig.AIP_COMPRESSION_ALGORITHM_TAR_GZIP: ("gzip", "tar.gzip"),
        ProcessingConfig.AIP_COMPRESSION_ALGORITHM_S7_COPY: ("7z", "copy"),
        ProcessingConfig.AIP_COMPRESSION_ALGORITHM_S7_BZIP2: ("7z", "bzip2"),
        ProcessingConfig.AIP_COMPRESSION_ALGORITHM_S7_LZMA: ("7z", "lzma"),
    }

    try:
        program, compression_algorithm = mapping[compression]
    except KeyError:
        msg = f"Invalid program-compression algorithm: {compression}"
        job.pyprint(msg, file=sys.stderr)
        return 255

    archive_path = f"{sip_name}-{sip_uuid}"
    uncompressed_location = sip_directory + archive_path

    # Even though no actual compression is taking place,
    # the location still needs to be set in the unit to ensure that the
    # %AIPFilename% variable is set appropriately.
    # Setting it to an empty string ensures the common
    # "%SIPDirectory%%AIPFilename%" pattern still points at the right thing.
    if program == "None":
        update_unit(sip_uuid, uncompressed_location)
        return 0

    job.pyprint(
        "Compressing {} with {}, algorithm {}, level {}".format(
            uncompressed_location, program, compression_algorithm, compression_level
        )
    )

    if program == "7z":
        compressed_location = uncompressed_location + ".7z"
        command = '/usr/bin/7z a -bd -t7z -y -m0={algorithm} -mx={level} -mta=on -mtc=on -mtm=on -mmt=on "{compressed_location}" "{uncompressed_location}"'.format(
            algorithm=compression_algorithm,
            level=compression_level,
            uncompressed_location=uncompressed_location,
            compressed_location=compressed_location,
        )
        tool_info_command = (
            r'echo program="7z"\; '
            r'algorithm="{}"\; '
            'version="`7z | grep Version`"'.format(compression_algorithm)
        )
    elif program == "pbzip2":
        compressed_location = uncompressed_location + ".tar.bz2"
        command = '/bin/tar -c --directory "{sip_directory}" "{archive_path}" | /usr/bin/pbzip2 --compress -{level} > "{compressed_location}"'.format(
            sip_directory=sip_directory,
            archive_path=archive_path,
            level=compression_level,
            compressed_location=compressed_location,
        )
        tool_info_command = (
            r'echo program="pbzip2"\; '
            r'algorithm="{}"\; '
            'version="$((pbzip2 -V) 2>&1)"'.format(compression_algorithm)
        )
    elif program == "gzip":
        compressed_location = uncompressed_location + ".tar.gz"
        command = '/bin/tar -c --directory "{sip_directory}" "{archive_path}" | /bin/gzip -{level} > "{compressed_location}"'.format(
            sip_directory=sip_directory,
            archive_path=archive_path,
            level=compression_level,
            compressed_location=compressed_location,
        )
        tool_info_command = (
            r'echo program="gzip"\; '
            r'algorithm="{}"\; '
            'version="$((gzip -V) 2>&1)"'.format(compression_algorithm)
        )
    else:
        msg = f"Program {program} not recognized, exiting script prematurely."
        job.pyprint(msg, file=sys.stderr)
        return 255

    job.pyprint("Executing command:", command)
    exit_code, std_out, std_err = executeOrRun(
        "bashScript", command, capture_output=True
    )
    job.write_output(std_out)
    job.write_error(std_err)

    # Add new AIP File
    file_uuid = sip_uuid
    databaseFunctions.insertIntoFiles(
        fileUUID=file_uuid,
        filePath=compressed_location.replace(sip_directory, "%SIPDirectory%", 1),
        sipUUID=sip_uuid,
        use="aip",
    )

    # Add compression event
    job.pyprint("Tool info command:", tool_info_command)
    _, tool_info, tool_info_err = executeOrRun(
        "bashScript", tool_info_command, capture_output=True
    )
    job.write_output(tool_info)
    job.write_error(tool_info_err)
    tool_output = f'Standard Output="{std_out}"; Standard Error="{std_err}"'
    databaseFunctions.insertIntoEvents(
        eventType="compression",
        eventDetail=tool_info,
        eventOutcomeDetailNote=tool_output,
        fileUUID=file_uuid,
    )

    update_unit(sip_uuid, compressed_location)

    return exit_code


def call(jobs):
    parser = argparse.ArgumentParser(description="Compress an AIP.")
    parser.add_argument("compression", type=str)
    parser.add_argument("compression_level", type=str)
    parser.add_argument("sip_directory", type=str, help="%SIPDirectory%")
    parser.add_argument("sip_name", type=str, help="%SIPName%")
    parser.add_argument("sip_uuid", type=str, help="%SIPUUID%")

    with transaction.atomic():
        for job in jobs:
            with job.JobContext():
                args = parser.parse_args(job.args[1:])
                job.set_status(
                    compress_aip(
                        job,
                        args.compression,
                        args.compression_level,
                        args.sip_directory,
                        args.sip_name,
                        args.sip_uuid,
                    )
                )



import logging
import os
import shutil
import sys
from pathlib import Path
from pprint import pformat

from bagit import Bag
from bagit import BagError
from django.conf import settings as mcpclient_settings
from django.db import transaction

from a3m import databaseFunctions
from a3m.executeOrRunSubProcess import executeOrRun
from a3m.main.models import File

logger = logging.getLogger(__name__)


class VerifyChecksumsError(Exception):
    """Checksum verification has failed."""


def extract_aip(job, aip_path, extract_path):
    os.makedirs(extract_path)
    command = ["atool", f"--extract-to={extract_path}", "-V0", aip_path]
    job.pyprint("Running extraction command:", command)
    exit_code, stdout, stderr = executeOrRun("command", command, capture_output=True)
    job.write_output(stdout)
    job.write_error(stderr)
    if exit_code != 0:
        raise Exception("Error extracting AIP")

    aip_identifier, ext = os.path.splitext(os.path.basename(aip_path))
    if ext in (".bz2", ".gz"):
        aip_identifier, _ = os.path.splitext(aip_identifier)
    return os.path.join(extract_path, aip_identifier)


def write_premis_event(
    job, sip_uuid, checksum_type, event_outcome, event_outcome_detail_note
):
    """Write the AIP-level "fixity check" PREMIS event."""
    try:
        databaseFunctions.insertIntoEvents(
            fileUUID=sip_uuid,
            eventType="fixity check",
            eventDetail='program="python, bag"; module="hashlib.{}()"'.format(
                checksum_type
            ),
            eventOutcome=event_outcome,
            eventOutcomeDetailNote=event_outcome_detail_note,
        )
    except Exception as err:
        job.pyprint(f"Failed to write PREMIS event to database. Error: {err}")
    else:
        return event_outcome_detail_note


def assert_checksum_types_match(job, file_, sip_uuid, settings_checksum_type):
    """Raise exception if checksum types (i.e., algorithms, e.g., 'sha256') of
    the file and the settings do not match.
    """
    if file_.checksumtype != settings_checksum_type:
        event_outcome_detail_note = (
            "The checksum type of file {file_uuid} is"
            " {file_checksum_type}; given the current application settings, we"
            " expect it to {settings_checksum_type}".format(
                file_uuid=file_.uuid,
                file_checksum_type=file_.checksumtype,
                settings_checksum_type=settings_checksum_type,
            )
        )
        raise VerifyChecksumsError(
            write_premis_event(
                job, sip_uuid, settings_checksum_type, "Fail", event_outcome_detail_note
            )
        )


def get_expected_checksum(job, bag, file_, sip_uuid, checksum_type, file_path):
    """Raise an exception if an expected checksum cannot be found in the
    Bag manifest.
    """
    try:
        return bag.entries[file_path][checksum_type]
    except KeyError:
        event_outcome_detail_note = (
            "Unable to find expected path {expected_path} for file"
            " {file_uuid} in the following mapping from file paths to"
            " checksums: {mapping}:".format(
                expected_path=file_path,
                file_uuid=file_.uuid,
                mapping=pformat(bag.entries),
            )
        )
        raise VerifyChecksumsError(
            write_premis_event(
                job, sip_uuid, checksum_type, "Fail", event_outcome_detail_note
            )
        )


def assert_checksums_match(job, file_, sip_uuid, checksum_type, expected_checksum):
    """Raise an exception if checksums do not match."""
    if file_.checksum != expected_checksum:
        event_outcome_detail_note = (
            "The checksum {db_checksum} for file {file_uuid} from the"
            " database did not match the corresponding checksum from the"
            " Bag manifest file {manifest_checksum}".format(
                file_uuid=file_.uuid,
                db_checksum=file_.checksum,
                manifest_checksum=expected_checksum,
            )
        )
        raise VerifyChecksumsError(
            write_premis_event(
                job, sip_uuid, checksum_type, "Fail", event_outcome_detail_note
            )
        )


def verify_checksums(job, bag, sip_uuid):
    """Verify that the checksums generated at the beginning of transfer match
    those generated near the end of ingest by bag, i.e., "Prepare AIP"
    (bagit_v0.0).
    """
    checksum_type = mcpclient_settings.DEFAULT_CHECKSUM_ALGORITHM
    removableFiles = [e.strip() for e in mcpclient_settings.REMOVABLE_FILES.split(",")]
    try:
        verification_count = 0
        for file_ in File.objects.filter(sip_id=sip_uuid):
            if (
                os.path.basename(file_.originallocation) in removableFiles
                or file_.removedtime
                or not file_.currentlocation.startswith("%SIPDirectory%objects/")
                or file_.filegrpuse == "manualNormalization"
            ):
                continue
            file_path = os.path.join(
                "data", file_.currentlocation.replace("%SIPDirectory%", "", 1)
            )
            assert_checksum_types_match(job, file_, sip_uuid, checksum_type)
            expected_checksum = get_expected_checksum(
                job, bag, file_, sip_uuid, checksum_type, file_path
            )
            assert_checksums_match(
                job, file_, sip_uuid, checksum_type, expected_checksum
            )
            verification_count += 1
    except VerifyChecksumsError as err:
        job.print_error(repr(err))
        raise
    event_outcome_detail_note = (
        "All checksums (count={verification_count}) generated at start of"
        " transfer match those generated by BagIt (bag).".format(
            verification_count=verification_count
        )
    )
    write_premis_event(job, sip_uuid, checksum_type, "Pass", event_outcome_detail_note)
    job.pyprint(event_outcome_detail_note)


def verify_aip(job):
    """Verify the AIP was bagged correctly by extracting it and running
    verification on its contents. This is also where we verify the checksums
    now that the verifyPREMISChecksums_v0.0 ("Verify checksums generated on
    ingest") micro-service has been removed. It was removed because verifying
    checksums by calculating them in that MS and then having bagit calculate
    them here was redundant.

    job.args[1] = UUID
      UUID of the SIP, which will become the UUID of the AIP
    job.args[2] = current location
      Full absolute path to the AIP's current location on the local filesystem
    """

    sip_uuid = job.args[1]  # %sip_uuid%
    aip_path = job.args[2]  # SIPDirectory%%sip_name%-%sip_uuid%.7z

    temp_dir = mcpclient_settings.TEMP_DIRECTORY

    is_uncompressed_aip = os.path.isdir(aip_path)

    if is_uncompressed_aip:
        bag_path = aip_path
    else:
        try:
            extract_dir = os.path.join(temp_dir, sip_uuid)
            bag_path = extract_aip(job, aip_path, extract_dir)
        except Exception as err:
            job.print_error(repr(err))
            job.pyprint(f'Error extracting AIP at "{aip_path}"', file=sys.stderr)
            return 1

    return_code = 0
    try:
        # Only validate completeness since we're going to verify checksums
        # later against what we have in the database via `verify_checksums`.
        bag = Bag(bag_path)
        bag.validate(completeness_only=True)
    except BagError as err:
        job.print_error(f"Error validating BagIt package: {err}")
        return_code = 1

    if return_code == 0:
        try:
            verify_checksums(job, bag, sip_uuid)
        except VerifyChecksumsError:
            return_code = 1
    else:
        job.pyprint(
            "Not verifying checksums because other tests have already" " failed."
        )

    # cleanup
    if not is_uncompressed_aip:
        try:
            shutil.rmtree(extract_dir)
        except OSError as err:
            job.pyprint(
                "Failed to remove temporary directory at {extract_dir} which"
                " contains the AIP extracted for verification."
                " Error:\n{err}".format(extract_dir=extract_dir, err=err),
                file=sys.stderr,
            )

    aip_path = Path(aip_path)
    completed_dir = Path(mcpclient_settings.SHARED_DIRECTORY, "completed")
    shutil.move(str(aip_path), str(completed_dir))
    logger.info("AIP generated: %s", aip_path.name)

    return return_code


def call(jobs):
    with transaction.atomic():
        for job in jobs:
            with job.JobContext():
                job.set_status(verify_aip(job))



from a3m.executeOrRunSubProcess import executeOrRun


def call(jobs):
    for job in jobs:
        with job.JobContext():
            exit_code, std_out, std_error = executeOrRun(
                "command", ["mkdir"] + job.args[1:], capture_output=True
            )

            job.write_error(std_error)
            job.write_output(std_out)
            job.set_status(exit_code)



# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: a3m/api/transferservice/v1beta1/service.proto
"""Generated protocol buffer code."""
from google.protobuf import descriptor as _descriptor
from google.protobuf import descriptor_pool as _descriptor_pool
from google.protobuf import symbol_database as _symbol_database
from google.protobuf.internal import builder as _builder

# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()


from a3m.api.transferservice.v1beta1 import (
    request_response_pb2 as a3m_dot_api_dot_transferservice_dot_v1beta1_dot_request__response__pb2,
)


DESCRIPTOR = _descriptor_pool.Default().AddSerializedFile(
    b'\n-a3m/api/transferservice/v1beta1/service.proto\x12\x1f\x61\x33m.api.transferservice.v1beta1\x1a\x36\x61\x33m/api/transferservice/v1beta1/request_response.proto2\xc5\x03\n\x0fTransferService\x12k\n\x06Submit\x12..a3m.api.transferservice.v1beta1.SubmitRequest\x1a/.a3m.api.transferservice.v1beta1.SubmitResponse"\x00\x12\x65\n\x04Read\x12,.a3m.api.transferservice.v1beta1.ReadRequest\x1a-.a3m.api.transferservice.v1beta1.ReadResponse"\x00\x12t\n\tListTasks\x12\x31.a3m.api.transferservice.v1beta1.ListTasksRequest\x1a\x32.a3m.api.transferservice.v1beta1.ListTasksResponse"\x00\x12h\n\x05\x45mpty\x12-.a3m.api.transferservice.v1beta1.EmptyRequest\x1a..a3m.api.transferservice.v1beta1.EmptyResponse"\x00\x42\xa9\x02\n#com.a3m.api.transferservice.v1beta1B\x0cServiceProtoP\x01ZUgithub.com/artefactual-labs/a3m/proto/a3m/api/transferservice/v1beta1;transferservice\xa2\x02\x03\x41\x41T\xaa\x02\x1f\x41\x33m.Api.Transferservice.V1beta1\xca\x02\x1f\x41\x33m\\Api\\Transferservice\\V1beta1\xe2\x02+A3m\\Api\\Transferservice\\V1beta1\\GPBMetadata\xea\x02"A3m::Api::Transferservice::V1beta1b\x06proto3'
)

_globals = globals()
_builder.BuildMessageAndEnumDescriptors(DESCRIPTOR, _globals)
_builder.BuildTopDescriptorsAndMessages(
    DESCRIPTOR, "a3m.api.transferservice.v1beta1.service_pb2", _globals
)
if _descriptor._USE_C_DESCRIPTORS == False:
    DESCRIPTOR._options = None
    DESCRIPTOR._serialized_options = b'\n#com.a3m.api.transferservice.v1beta1B\014ServiceProtoP\001ZUgithub.com/artefactual-labs/a3m/proto/a3m/api/transferservice/v1beta1;transferservice\242\002\003AAT\252\002\037A3m.Api.Transferservice.V1beta1\312\002\037A3m\\Api\\Transferservice\\V1beta1\342\002+A3m\\Api\\Transferservice\\V1beta1\\GPBMetadata\352\002"A3m::Api::Transferservice::V1beta1'
    _globals["_TRANSFERSERVICE"]._serialized_start = 139
    _globals["_TRANSFERSERVICE"]._serialized_end = 592
# @@protoc_insertion_point(module_scope)



"""
Exposes various metrics via Prometheus.
"""
import functools
import importlib
import pkgutil

from django.conf import settings
from django.db.models import Sum
from django.utils import timezone
from prometheus_client import Counter
from prometheus_client import Gauge
from prometheus_client import Histogram

from a3m.client import clientScripts
from a3m.common_metrics import PACKAGE_FILE_COUNT_BUCKETS
from a3m.common_metrics import PACKAGE_SIZE_BUCKETS
from a3m.common_metrics import PROCESSING_TIME_BUCKETS
from a3m.common_metrics import TASK_DURATION_BUCKETS
from a3m.main.models import File
from a3m.main.models import Transfer


job_counter = Counter(
    "mcpclient_job_total",
    "Number of jobs processed, labeled by script",
    ["script_name"],
)
job_processed_timestamp = Gauge(
    "mcpclient_job_success_timestamp",
    "Timestamp of most recent job processed, labeled by script",
    ["script_name"],
)
job_error_counter = Counter(
    "mcpclient_job_error_total",
    "Number of failures processing jobs, labeled by script",
    ["script_name"],
)
job_error_timestamp = Gauge(
    "mcpclient_job_error_timestamp",
    "Timestamp of most recent job failure, labeled by script",
    ["script_name"],
)

task_execution_time_histogram = Histogram(
    "mcpclient_task_execution_time_seconds",
    "Histogram of worker task execution times in seconds, labeled by script",
    ["script_name"],
    buckets=TASK_DURATION_BUCKETS,
)

transfer_started_counter = Counter(
    "mcpclient_transfer_started_total", "Number of Transfers started"
)
transfer_started_timestamp = Gauge(
    "mcpclient_transfer_started_timestamp", "Timestamp of most recent transfer started"
)
transfer_completed_counter = Counter(
    "mcpclient_transfer_completed_total", "Number of Transfers completed"
)
transfer_completed_timestamp = Gauge(
    "mcpclient_transfer_completed_timestamp",
    "Timestamp of most recent transfer completed",
)
transfer_error_counter = Counter(
    "mcpclient_transfer_error_total",
    "Number of transfer failures, by error type",
    ["failure_type"],
)
transfer_error_timestamp = Gauge(
    "mcpclient_transfer_error_timestamp",
    "Timestamp of most recent transfer failure, by error type",
    ["failure_type"],
)
transfer_files_histogram = Histogram(
    "mcpclient_transfer_files",
    "Histogram of number of files included in transfers",
    buckets=PACKAGE_FILE_COUNT_BUCKETS,
)
transfer_size_histogram = Histogram(
    "mcpclient_transfer_size_bytes",
    "Histogram of number bytes in transfers",
    buckets=PACKAGE_SIZE_BUCKETS,
)

sip_started_counter = Counter("mcpclient_sip_started_total", "Number of SIPs started")
sip_started_timestamp = Gauge(
    "mcpclient_sip_started_timestamp", "Timestamp of most recent SIP started"
)
sip_error_counter = Counter(
    "mcpclient_sip_error_total",
    "Number of SIP failures, by error type",
    ["failure_type"],
)
sip_error_timestamp = Gauge(
    "mcpclient_sip_error_timestamp",
    "Timestamp of most recent SIP failure, by error type",
    ["failure_type"],
)

aips_stored_counter = Counter("mcpclient_aips_stored_total", "Number of AIPs stored")
aips_stored_timestamp = Gauge(
    "mcpclient_aips_stored_timestamp", "Timestamp of most recent AIP stored"
)
aip_processing_time_histogram = Histogram(
    "mcpclient_aip_processing_seconds",
    "Histogram of AIP processing time, from first file recorded in DB to storage in SS",
    buckets=PROCESSING_TIME_BUCKETS,
)
aip_files_stored_histogram = Histogram(
    "mcpclient_aip_files_stored",
    "Histogram of number of files stored in AIPs. Note, this includes metadata, derivatives, etc.",
    buckets=PACKAGE_FILE_COUNT_BUCKETS,
)
aip_size_histogram = Histogram(
    "mcpclient_aip_size_bytes",
    "Histogram of number of bytes stored in AIPs. Note, this includes metadata, derivatives, etc.",
    buckets=PACKAGE_SIZE_BUCKETS,
)

# There's no central place to pull these constants from currently
PACKAGE_FAILURE_TYPES = ("fail", "reject")


def skip_if_prometheus_disabled(func):
    @functools.wraps(func)
    def wrapper(*args, **kwds):
        if settings.PROMETHEUS_ENABLED:
            return func(*args, **kwds)
        return None

    return wrapper


def init_counter_labels():
    # Zero our counters to start, by intializing all labels. Non-zero starting points
    # cause problems when measuring rates.

    for _, modname, is_pkg in pkgutil.iter_modules(clientScripts.__path__):
        if not is_pkg:
            continue
        module = importlib.import_module(modname)
        if "call" not in module:
            raise Exception()
        modname = modname.split(".")[-1]
        print(modname)
        job_counter.labels(script_name=modname)
        job_processed_timestamp.labels(script_name=modname)
        job_error_counter.labels(script_name=modname)
        job_error_timestamp.labels(script_name=modname)
        task_execution_time_histogram.labels(script_name=modname)

    for failure_type in PACKAGE_FAILURE_TYPES:
        transfer_error_counter.labels(failure_type=failure_type)
        transfer_error_timestamp.labels(failure_type=failure_type)

    for failure_type in PACKAGE_FAILURE_TYPES:
        sip_error_counter.labels(failure_type=failure_type)
        sip_error_timestamp.labels(failure_type=failure_type)


@skip_if_prometheus_disabled
def job_completed(script_name):
    job_counter.labels(script_name=script_name).inc()
    job_processed_timestamp.labels(script_name=script_name).set_to_current_time()


@skip_if_prometheus_disabled
def job_failed(script_name):
    job_counter.labels(script_name=script_name).inc()
    job_error_counter.labels(script_name=script_name).inc()
    job_error_timestamp.labels(script_name=script_name).set_to_current_time()


@skip_if_prometheus_disabled
def aip_stored(sip_uuid, size):
    aips_stored_counter.inc()
    aips_stored_timestamp.set_to_current_time()
    aip_size_histogram.observe(size)

    try:
        earliest_file = File.objects.filter(sip_id=sip_uuid).earliest("enteredsystem")
    except File.DoesNotExist:
        pass
    else:
        duration = (timezone.now() - earliest_file.enteredsystem).total_seconds()
        aip_processing_time_histogram.observe(duration)

    # We do two queries here, as we may not have format information for everything
    total_file_count = File.objects.filter(sip_id=sip_uuid).count()
    aip_files_stored_histogram.observe(total_file_count)


@skip_if_prometheus_disabled
def transfer_started():
    transfer_started_counter.inc()
    transfer_started_timestamp.set_to_current_time()


@skip_if_prometheus_disabled
def transfer_completed(transfer_uuid):
    try:
        transfer = Transfer.objects.get(uuid=transfer_uuid)
    except Transfer.DoesNotExist:
        return

    transfer_completed_counter.inc()
    transfer_completed_timestamp.set_to_current_time()

    file_queryset = File.objects.filter(transfer=transfer)
    file_count = file_queryset.count()
    transfer_files_histogram.observe(file_count)

    transfer_size = file_queryset.aggregate(total_size=Sum("size"))
    transfer_size_histogram.observe(transfer_size["total_size"])


@skip_if_prometheus_disabled
def transfer_failed(failure_type):
    transfer_error_counter.labels(failure_type=failure_type).inc()
    transfer_error_timestamp.labels(failure_type=failure_type).set_to_current_time()


@skip_if_prometheus_disabled
def sip_started():
    sip_started_counter.inc()
    sip_started_timestamp.set_to_current_time()


@skip_if_prometheus_disabled
def sip_failed(failure_type):
    sip_error_counter.labels(failure_type=failure_type).inc()
    sip_error_timestamp.labels(failure_type=failure_type).set_to_current_time()



"""Workflow decoder and validator.

The main function to start working with this module is ``load``. It decodes the
JSON-encoded bytes and validates the document against the schema.

    >>> import workflow
    >>> with open("workflow.json") as file_object:
            wf = workflow.load(file_object)

If the document cannot be validated, ``jsonschema.ValidationError`` is raised.
Otherwise, ``load`` will return an instance of ``Workflow`` which is used in
MCPServer to read workflow links that can be instances of three different
classes ``Chain``, ``Link`` and ``WatchedDir``. They have different method
sets.
"""
import json
import os

from jsonschema import FormatChecker
from jsonschema import validate
from jsonschema.exceptions import ValidationError

from a3m.server.jobs import Job
from a3m.server.translation import FALLBACK_LANG
from a3m.server.translation import TranslationLabel


_LATEST_SCHEMA = "workflow-schema-v1.json"
ASSETS_DIR = os.path.join(
    os.path.dirname(os.path.dirname(os.path.abspath(os.path.join(__file__)))), "assets"
)

DEFAULT_WORKFLOW = os.path.join(ASSETS_DIR, "workflow.json")


def _invert_job_statuses():
    """Return an inverted dict of job statuses, i.e. indexed by labels."""
    statuses = {}
    for status in Job.STATUSES:
        label = str(status[1])
        statuses[label] = status[0]

    return statuses


# Job statuses (from ``Job.STATUSES``) indexed by the English labels.
# This is useful when decoding the values used in the JSON-encoded workflow
# where we're using labels instead of IDs.
_STATUSES = _invert_job_statuses()


class Workflow:
    def __init__(self, parsed_obj):
        self._src = parsed_obj
        self._decode_links()

    def __str__(self):
        return f"Links {len(self.links)}"

    def _decode_links(self):
        self.links = {}
        for link_id, link_obj in self._src["links"].items():
            self.links[link_id] = Link(link_id, link_obj, self)

    def get_links(self):
        return self.links

    def get_link(self, link_id):
        return self.links[link_id]

    def get_initiator(self):
        for _, link in self.links.items():
            if link.is_initiator:
                return link


class BaseLink:
    def __str__(self):
        return self.id

    def get_label(self, key, lang=FALLBACK_LANG, fallback_label=None):
        """Proxy to find translated attributes."""
        try:
            instance = self._src[key]
        except KeyError:
            return None
        return instance.get_label(lang, fallback_label)

    def _decode_translation(self, translation_dict):
        return TranslationLabel(translation_dict)

    @property
    def workflow(self):
        return self._workflow


class Link(BaseLink):
    def __init__(self, id_, attrs, workflow):
        self.id = id_
        self._src = attrs
        self._workflow = workflow
        self._decode_job_statuses()
        self._decode_translations()

    def __repr__(self):
        return f"Link <{self.id}>"

    def __getitem__(self, key):
        return self._src[key]

    def _decode_job_statuses(self):
        """Replace status labels with their IDs.

        In JSON, a job status is encoded using its English label, e.g. "Failed"
        instead of the corresponding value in ``JOB.STATUS_FAILED``. This
        method decodes the statuses so it becomes easier to work with them
        internally.
        """
        self._src["fallback_job_status"] = _STATUSES[self._src["fallback_job_status"]]
        for obj in self.exit_codes.values():
            obj["job_status"] = _STATUSES[obj["job_status"]]

    def _decode_translations(self):
        self._src["description"] = self._decode_translation(self._src["description"])
        self._src["group"] = self._decode_translation(self._src["group"])
        config = self._src["config"]
        if config["@manager"] == "linkTaskManagerReplacementDicFromChoice":
            for item in config["replacements"]:
                item["description"] = self._decode_translation(item["description"])

    @property
    def config(self):
        return self._src["config"]

    @property
    def is_initiator(self):
        """Check if the link is indicated as an initiator link."""
        return self._src.get("start", False)

    @property
    def is_terminal(self):
        """Check if the link is indicated as a terminal link."""
        return self._src.get("end", False)

    @property
    def exit_codes(self):
        return self._src.get("exit_codes", {})

    def get_next_link(self, code):
        code = str(code)
        try:
            link_id = self.exit_codes[code].get("link_id")
        except KeyError:
            link_id = self._src["fallback_link_id"]
        return self._workflow.get_link(link_id)

    def get_status_id(self, code):
        """Return the expected Job status ID given an exit code."""
        code = str(code)
        try:
            status_id = self.exit_codes[code]["job_status"]
        except KeyError:
            status_id = self._src["fallback_job_status"]
        return status_id


class WorkflowJSONDecoder(json.JSONDecoder):
    def decode(self, foo, **kwargs):
        parsed_json = super().decode(foo, **kwargs)
        return Workflow(parsed_json)


def load(fp):
    """Read JSON document from file-like object, validate and decode it."""
    blob = fp.read()  # Read once, used twice.
    _validate(blob)
    parsed = json.loads(blob, cls=WorkflowJSONDecoder)

    return parsed


def load_default_workflow() -> Workflow:
    with open(DEFAULT_WORKFLOW) as default_workflow:
        return load(default_workflow)


class SchemaValidationError(ValidationError):
    """It wraps ``jsonschema.exceptions.ValidationError``."""


def _validate(blob):
    """Decode and validate the JSON document."""
    try:
        validate(json.loads(blob), _get_schema(), format_checker=FormatChecker())
    except ValidationError as err:
        raise SchemaValidationError(**err._contents())


def _get_schema():
    """Decode the default schema and return it."""
    schema = os.path.join(ASSETS_DIR, _LATEST_SCHEMA)
    with open(schema) as fp:
        return json.load(fp)



"""
A job corresponds to a microservice, a link in the workflow, and the `Jobs`
table in the database.

Initialization of `Job` objects is typically done via a `JobChain`,
corresponding to a chain in the workflow. The `JobChain` object handles
determining the next job to be run, and passing data between jobs.

The `Job` class is a base class for other job types. There are various
concrete types of jobs, handled by subclasses:
    * `ClientScriptJob`, handling Jobs to be execute on MCPClient
    * `NextLinkDecisionJob`, handling workflow decision points
"""
from a3m.server.jobs.base import Job
from a3m.server.jobs.chain import JobChain
from a3m.server.jobs.client import ClientScriptJob
from a3m.server.jobs.client import DirectoryClientScriptJob
from a3m.server.jobs.client import FilesClientScriptJob
from a3m.server.jobs.decisions import NextLinkDecisionJob

__all__ = (
    "ClientScriptJob",
    "DirectoryClientScriptJob",
    "FilesClientScriptJob",
    "Job",
    "JobChain",
    "NextLinkDecisionJob",
)



import datetime
import logging
import os
import sys
import uuid

import metsrw
from django.db import transaction
from django.utils import dateparse
from django.utils import timezone
from lxml import etree

from a3m.main.models import Agent
from a3m.main.models import Event
from a3m.main.models import File

logger = logging.getLogger(__name__)
FAILURE = 1
SUCCESS = 0
TRANSFER_ORIGINAL_LOCATION_PREFIX = r"%transferDirectory%"


def log_missing_original_name(file_identifier, printfn=print):
    printfn(
        "The {formatted_object_identifier} does not contain a premis:originalName element".format(
            formatted_object_identifier=format_identifier(
                file_identifier, element_type="object"
            )
        ),
        file=sys.stderr,
    )


def log_filename_mismatch(file_identifier, original_name, printfn=print):
    printfn(
        "The {formatted_object_identifier} contains a premis:originalName value that does not match any filename in this transfer: '{original_name}'".format(
            formatted_object_identifier=format_identifier(
                file_identifier, element_type="object"
            ),
            original_name=original_name,
        ),
        file=sys.stderr,
    )


def log_missing_xsd(path, printfn=print):
    printfn(f"The PREMIS XML schema asset {path} is unavailable", file=sys.stderr)


def log_invalid_xsd(path, exception, printfn=print):
    printfn(
        f"Could not parse the PREMIS XML schema {path}", str(exception), file=sys.stderr
    )


def log_invalid_events_xml(path, exception, printfn=print):
    printfn(
        "Could not validate the events XML file {} using the PREMIS 3 XML schema".format(
            path
        ),
        str(exception),
        file=sys.stderr,
    )


def log_event_without_files(event_identifier, printfn=print):
    printfn(
        "The {formatted_event_identifier} is not related to any filename in the transfer files".format(
            formatted_event_identifier=format_identifier(
                event_identifier, element_type="event"
            )
        ),
        file=sys.stderr,
    )


def log_event_without_agents(event_identifier, printfn=print):
    printfn(
        "The {formatted_event_identifier} is not related to any agents".format(
            formatted_event_identifier=format_identifier(
                event_identifier, element_type="event"
            )
        ),
        file=sys.stderr,
    )


def log_missing_events_xml(path, printfn=print):
    printfn(f"No events XML file found at path {path}")


def log_invalid_event_datetime(premis_element, printfn=print):
    printfn(
        "The {formatted_event_identifier} contains a premis:eventDateTime value that could not be parsed: {event_datetime}".format(
            formatted_event_identifier=format_identifier(
                get_identifier(premis_element), element_type="event"
            ),
            event_datetime=premis_element.date_time,
        ),
        file=sys.stderr,
    )


def log_unrelated_file(file_identifier, printfn=print):
    printfn(
        "The {formatted_object_identifier} is not related to any premis:event element".format(
            formatted_object_identifier=format_identifier(
                file_identifier, element_type="object"
            )
        ),
        file=sys.stderr,
    )


def log_unrelated_agent(agent_identifier, printfn=print):
    printfn(
        "The {formatted_agent_identifier} is not related to any premis:event element".format(
            formatted_agent_identifier=format_identifier(
                agent_identifier, element_type="agent"
            )
        ),
        file=sys.stderr,
    )


def log_unrelated_event_to_files(event_identifier, printfn=print):
    printfn(
        "The {formatted_event_identifier} is not related to any premis:object element".format(
            formatted_event_identifier=format_identifier(
                event_identifier, element_type="event"
            )
        ),
        file=sys.stderr,
    )


def log_unrelated_event_to_agents(event_identifier, printfn=print):
    printfn(
        "The {formatted_event_identifier} is not related to any premis:agent element".format(
            formatted_event_identifier=format_identifier(
                event_identifier, element_type="event"
            )
        ),
        file=sys.stderr,
    )


def log_file_related_to_nonexistent_events(
    file_identifier, nonexistent_identifiers, printfn=print
):
    printfn(
        "The {formatted_object_identifier} contains premis:linkingEventIdentifier elements that reference premis:event elements that are not present in this document: {formatted_event_identifiers}".format(
            formatted_object_identifier=format_identifier(
                file_identifier, element_type="object"
            ),
            formatted_event_identifiers=format_as_grouped_identifiers(
                nonexistent_identifiers, element_type="event"
            ),
        ),
        file=sys.stderr,
    )


def log_agent_related_to_nonexistent_events(
    agent_identifier, nonexistent_identifiers, printfn=print
):
    printfn(
        "The {formatted_agent_identifier} contains premis:linkingEventIdentifier elements that reference premis:event elements that are not present in this document: {formatted_event_identifiers}".format(
            formatted_agent_identifier=format_identifier(
                agent_identifier, element_type="agent"
            ),
            formatted_event_identifiers=format_as_grouped_identifiers(
                nonexistent_identifiers, element_type="event"
            ),
        ),
        file=sys.stderr,
    )


def log_event_related_to_nonexistent_files(
    event_identifier, nonexistent_identifiers, printfn=print
):
    printfn(
        "The {formatted_event_identifier} contains premis:linkingObjectIdentifier elements that reference premis:object elements that are not present in this document: {formatted_object_identifiers}".format(
            formatted_event_identifier=format_identifier(
                event_identifier, element_type="event"
            ),
            formatted_object_identifiers=format_as_grouped_identifiers(
                nonexistent_identifiers, element_type="object"
            ),
        ),
        file=sys.stderr,
    )


def log_event_related_to_nonexistent_agents(
    event_identifier, nonexistent_identifiers, printfn
):
    printfn(
        "The {formatted_event_identifier} contains premis:linkingAgentIdentifier elements that reference premis:agent elements that are not present in this document: {formatted_agent_identifiers}".format(
            formatted_event_identifier=format_identifier(
                event_identifier, element_type="event"
            ),
            formatted_agent_identifiers=format_as_grouped_identifiers(
                nonexistent_identifiers, element_type="agent"
            ),
        ),
        file=sys.stderr,
    )


def log_event_id_change(event_id, new_event_id, printfn=print):
    printfn(f"Changed event identifier from {event_id} to {new_event_id}")


def log_event_successfully_created(db_event, printfn=print):
    printfn(
        "Imported PREMIS {} event and assigned identifier {}".format(
            db_event.event_type, db_event.event_id
        )
    )


class PREMISFile(metsrw.plugins.premisrw.PREMISObject):
    """
    Implements the originalName child element used to relate files in a
    transfer to events.
    """

    @property
    def schema(self):
        object_schema = super().schema
        return object_schema + (("original_name",),)


def parse_datetime(value):
    parsed_datetime = dateparse.parse_datetime(value)
    # Check if we got a date
    if parsed_datetime is None:
        parsed_date = dateparse.parse_date(value)
        if parsed_date is not None:
            parsed_datetime = datetime.datetime.combine(
                parsed_date, datetime.datetime.min.time()
            )
    if value and parsed_datetime is None:
        raise ValueError(f"Unable to parse '{value}' as a datetime")
    # If we didn't get a timezone, use the one configured in Django settings
    if parsed_datetime is not None and parsed_datetime.tzinfo is None:
        parsed_datetime = parsed_datetime.replace(
            tzinfo=timezone.get_default_timezone()
        )
    return parsed_datetime


def get_premis_schema(premis_xsd_path, printfn=print):
    if not os.path.isfile(premis_xsd_path):
        log_missing_xsd(premis_xsd_path, printfn)
        return
    try:
        return etree.XMLSchema(
            etree.parse(  # noqa S320
                premis_xsd_path,
                etree.XMLParser(resolve_entities=False, no_network=True),
            )
        )
    except (etree.XMLSyntaxError, etree.XMLSchemaParseError) as exception:
        log_invalid_xsd(premis_xsd_path, exception, printfn)
        return


def get_validated_etree(premis_events_xml_path, premis_schema, printfn=print):
    if not os.path.isfile(premis_events_xml_path):
        # a missing events XML file is not considered an error
        log_missing_events_xml(premis_events_xml_path, printfn)
        return
    try:
        result = etree.parse(  # noqa S320
            premis_events_xml_path,
            etree.XMLParser(resolve_entities=False, no_network=True),
        )
        premis_schema.assertValid(result)
        return result
    except (etree.XMLSyntaxError, etree.DocumentInvalid) as exception:
        log_invalid_events_xml(premis_events_xml_path, exception, printfn)
        return


def get_xml_tree(xsd_path, xml_path, printfn=print):
    schema = get_premis_schema(xsd_path, printfn)
    if schema is None:
        return None
    return get_validated_etree(xml_path, schema, printfn)


def get_elements(tree, selector, element_factory, printfn=print):
    """
    Looks for PREMISElements based on a selector and builds a dictionary
    keyed by identifier that contains element attributes.
    """
    result = {}
    root = tree.getroot()
    premis_version = root.get("version", metsrw.plugins.premisrw.PREMIS_VERSION)
    elements = tree.findall(
        selector,
        metsrw.plugins.premisrw.PREMIS_VERSIONS_MAP[premis_version]["namespaces"],
    )
    # metsrw.plugins.premisrw.premis_to_data (used later) expects the PREMISElements
    # to have their own version attribute, so we set it if they don't
    for element in elements:
        if element.get("version") is None:
            element.set("version", premis_version)
        attributes = element_factory(element, printfn)
        result[attributes["identifier"]] = attributes
    return result


def format_identifier_details(identifier, element_type):
    identifier_type = identifier[0]
    identifier_value = identifier[1]
    return "premis:{element_type}IdentifierType='{identifier_type}' and premis:{element_type}IdentifierValue='{identifier_value}'".format(
        identifier_type=identifier_type,
        identifier_value=identifier_value,
        element_type=element_type,
    )


def format_identifier(identifier, element_type):
    return "premis:{element_type} element with {formatted_identifier_details}".format(
        element_type=element_type,
        formatted_identifier_details=format_identifier_details(
            identifier, element_type=element_type
        ),
    )


def format_as_grouped_identifiers(identifiers, element_type):
    return ",".join(
        [
            "({formatted_identifier_details})".format(
                formatted_identifier_details=format_identifier_details(
                    identifier, element_type=element_type
                )
            )
            for identifier in identifiers
        ]
    )


def get_premis_element_children_identifiers(premis_element, selector):
    result = set()
    for identifier in premis_element.findall(selector):
        result.add((identifier.type, identifier.value))
    return result


def get_premis_element_children_text_attribute(
    premis_element, selector, attribute_name
):
    """
    Get children of a PREMISElement based on a selector. If the children contain
    text, they're joined as a single string.
    """
    return (
        " ".join(
            [
                getattr(e, attribute_name)
                for e in premis_element.findall(selector)
                if isinstance(getattr(e, attribute_name), str)
            ]
        )
        or None
    )


def get_identifier(premis_element):
    return (premis_element.identifier_type, premis_element.identifier_value)


def file_element_factory(element, printfn=print):
    premis_element = PREMISFile(data=metsrw.plugins.premisrw.premis_to_data(element))
    result = {"identifier": get_identifier(premis_element)}
    if isinstance(premis_element.original_name, str):
        result["original_name"] = premis_element.original_name.strip()
    else:
        result["original_name"] = ""
    result["events"] = get_premis_element_children_identifiers(
        premis_element, "linking_event_identifier"
    )
    return result


def agent_element_factory(element, printfn=print):
    premis_element = metsrw.plugins.premisrw.PREMISAgent(
        data=metsrw.plugins.premisrw.premis_to_data(element)
    )
    result = {
        "identifier": get_identifier(premis_element),
        "name": premis_element.name,
        "type": premis_element.type,
    }
    result["events"] = get_premis_element_children_identifiers(
        premis_element, "linking_event_identifier"
    )
    return result


def event_element_factory(element, printfn=print):
    premis_element = metsrw.plugins.premisrw.PREMISEvent(
        data=metsrw.plugins.premisrw.premis_to_data(element)
    )
    result = {
        "identifier": get_identifier(premis_element),
        "event_id": premis_element.identifier_value,
        "event_type": premis_element.type,
    }
    event_datetime = None
    if premis_element.date_time:
        try:
            event_datetime = parse_datetime(premis_element.date_time)
        except ValueError:
            log_invalid_event_datetime(premis_element, printfn)
    if event_datetime is not None:
        result["event_datetime"] = event_datetime
    event_detail = get_premis_element_children_text_attribute(
        premis_element, "event_detail_information", "event_detail"
    )
    if event_detail:
        result["event_detail"] = event_detail
    event_outcome = get_premis_element_children_text_attribute(
        premis_element, "event_outcome_information", "event_outcome"
    )
    if event_outcome:
        result["event_outcome"] = event_outcome
    event_outcome_detail = get_premis_element_children_text_attribute(
        premis_element,
        "event_outcome_information/event_outcome_detail",
        "event_outcome_detail_note",
    )
    if event_outcome_detail:
        result["event_outcome_detail"] = event_outcome_detail
    result["files"] = get_premis_element_children_identifiers(
        premis_element, "linking_object_identifier"
    )
    result["agents"] = get_premis_element_children_identifiers(
        premis_element, "linking_agent_identifier"
    )
    return result


def get_invalid_file_identifiers(files, file_queryset, printfn=print):
    """
    File dictionaries are considered invalid if their `original_name` is
    empty or if its filename is not found in the transfer's file queryset.
    """
    result = set()
    for file_identifier, file_ in files.items():
        original_name = file_["original_name"]
        if not original_name:
            log_missing_original_name(file_identifier, printfn)
            result.add(file_identifier)
            continue
        original_location = "".join([TRANSFER_ORIGINAL_LOCATION_PREFIX, original_name])
        if not file_queryset.filter(originallocation=original_location).exists():
            log_filename_mismatch(file_identifier, original_name, printfn)
            result.add(file_identifier)
    return result


def print_unrelated_files(files, events, printfn=print):
    """
    A file dictionary is considered unrelated if its key that corresponds to
    its identifier doesn't appear in the `files` set of any event.

    Return a boolean indicating if there are unrelated file elements.
    """
    result = False
    related_files = set()
    for event in events.values():
        related_files.update(event["files"])
    for file_identifier in files:
        if file_identifier not in related_files:
            log_unrelated_file(file_identifier, printfn)
            result = True
    return result


def print_unrelated_agents(agents, events, printfn=print):
    """
    An agent dictionary is considered unrelated if its key that corresponds to
    its identifier doesn't appear in the `agents` set of any event.

    Return a boolean indicating if there are unrelated agent elements.
    """
    result = False
    related_agents = set()
    for event in events.values():
        related_agents.update(event["agents"])
    for agent_identifier in agents:
        if agent_identifier not in related_agents:
            log_unrelated_agent(agent_identifier, printfn)
            result = True
    return result


def print_unrelated_events(events, printfn=print):
    """
    An event dictionary is considered unrelated if it doesn't have any identifiers
    in its `files` or `agents` sets.

    Return a boolean indicating if there are unrelated event elements.
    """
    result = False
    for event_identifier, event in events.items():
        if not event["files"]:
            log_unrelated_event_to_files(event_identifier, printfn)
            result = True
        if not event["agents"]:
            log_unrelated_event_to_agents(event_identifier, printfn)
            result = True
    return result


def print_unrelated_elements(files, agents, events, printfn=print):
    """
    Go through each relationship and return a boolean indicating if any unrelated
    element exists.
    """
    return (
        print_unrelated_files(files, events, printfn)
        or print_unrelated_agents(agents, events, printfn)
        or print_unrelated_events(events, printfn)
    )


def print_files_related_to_nonexistent_events(files, events, printfn=print):
    """
    Search the events set of each file and print event identifiers that don't
    exist in the events collection.

    Return a boolean indicating if there are any file elements referencing
    nonexistent events.
    """
    result = False
    event_identifiers = set(events)
    for file_identifier, file_ in files.items():
        nonexistent_identifiers = file_["events"].difference(event_identifiers)
        if nonexistent_identifiers:
            log_file_related_to_nonexistent_events(
                file_identifier, nonexistent_identifiers, printfn
            )
            result = True
    return result


def print_agents_related_to_nonexistent_events(agents, events, printfn=print):
    """
    Search the events set of each agent and print event identifiers that don't
    exist in the events collection.

    Return a boolean indicating if there are any agent elements referencing
    nonexistent events.
    """
    result = False
    event_identifiers = set(events)
    for agent_identifier, agent in agents.items():
        nonexistent_identifiers = agent["events"].difference(event_identifiers)
        if nonexistent_identifiers:
            log_agent_related_to_nonexistent_events(
                agent_identifier, nonexistent_identifiers, printfn
            )
            result = True
    return result


def print_events_related_to_nonexistent_files(events, files, printfn=print):
    """
    Search the files set of each event and print event identifiers that don't
    exist in the events collection.

    Return a boolean indicating if there are any event elements referencing
    nonexistent files.
    """
    result = False
    file_identifiers = set(files)
    for event_identifier, event in events.items():
        nonexistent_identifiers = event["files"].difference(file_identifiers)
        if nonexistent_identifiers:
            log_event_related_to_nonexistent_files(
                event_identifier, nonexistent_identifiers, printfn
            )
            result = True
    return result


def print_events_related_to_nonexistent_agents(events, agents, printfn=print):
    """
    Search the agents set of each event and print event identifiers that don't
    exist in the events collection.

    Return a boolean indicating if there are any event elements referencing
    nonexistent agents.
    """
    result = False
    agent_identifiers = set(agents)
    for event_identifier, event in events.items():
        nonexistent_identifiers = event["agents"].difference(agent_identifiers)
        if nonexistent_identifiers:
            log_event_related_to_nonexistent_agents(
                event_identifier, nonexistent_identifiers, printfn
            )
            result = True
    return result


def print_nonexistent_references(files, agents, events, printfn=print):
    """
    Go through each relationship and return a boolean indicating if an element
    references any nonexistent element.
    """
    return (
        print_files_related_to_nonexistent_events(files, events, printfn)
        or print_agents_related_to_nonexistent_events(agents, events, printfn)
        or print_events_related_to_nonexistent_files(events, files, printfn)
        or print_events_related_to_nonexistent_agents(events, agents, printfn)
    )


def get_or_create_agents(agents):
    """
    Get a list of Agent model instances from each agent dictionary.
    """
    result = []
    for agent in agents:
        db_agent, _ = Agent.objects.get_or_create(
            identifiertype=agent["identifier"][0],
            identifiervalue=agent["identifier"][1],
            name=agent["name"],
            agenttype=agent["type"],
        )
        result.append(db_agent)
    return result


def get_event_agents(event, agents, agent_identifiers, printfn=print):
    """
    Get agents for the event ensuring that they exist in the agents collection.

    A message identifying the event is logged if no agents can be returned.
    """
    # get only agent identifiers that exist in the agents collection
    existent_agent_identifiers = event["agents"].intersection(agent_identifiers)
    return [agents[agent_identifier] for agent_identifier in existent_agent_identifiers]


def get_event_files(
    event, files, file_identifiers, file_identifiers_to_ignore, printfn=print
):
    """
    Get files for the event ensuring that they exist in the files collection.
    Invalid files are ignored.

    A message identifying the event is logged if no events can be returned.
    """
    # get only file identifiers that exist in the files collection
    event_file_identifiers = event["files"].intersection(file_identifiers)
    # and ignore invalid files
    event_file_identifiers = event_file_identifiers.difference(
        file_identifiers_to_ignore
    )
    return [files[file_identifier] for file_identifier in event_file_identifiers]


def ensure_event_id_is_uuid(event_id, printfn=print):
    try:
        uuid.UUID(event_id, version=4)
        if Event.objects.filter(event_id=event_id).exists():
            raise ValueError(f"There is already an event with this event_id {event_id}")
    except ValueError:
        new_event_id = uuid.uuid4()
        log_event_id_change(event_id, new_event_id, printfn)
        event_id = new_event_id
    return event_id


def save_events(valid_events, file_queryset, printfn=print):
    for valid_event in valid_events:
        event = valid_event["event"]

        db_agents = get_or_create_agents(valid_event["event_agents"])

        for file_ in valid_event["event_files"]:
            # get database file from originalName
            db_file = file_queryset.get(
                originallocation="".join(
                    [TRANSFER_ORIGINAL_LOCATION_PREFIX, file_["original_name"]]
                )
            )

            # ensure the event identifier is uuid
            event["event_id"] = ensure_event_id_is_uuid(event["event_id"], printfn)

            # get only attributes suitable for the database model
            event_attributes = event.copy()
            del event_attributes["identifier"]
            del event_attributes["files"]
            del event_attributes["agents"]

            # create database event
            db_event = Event.objects.create(file_uuid=db_file, **event_attributes)
            db_event.agents.add(*db_agents)

            # The event_datetime field is auto_now, which will ignore what we pass
            # to create. Work around this with another query to update the row :(
            if event_attributes.get("event_datetime") is not None:
                Event.objects.filter(pk=db_event.pk).update(
                    event_datetime=event_attributes["event_datetime"]
                )

            # print success message
            log_event_successfully_created(db_event, printfn)


def relate_files_to_events(files, events):
    """
    Ensure the identifier of a file is included in the events collection
    for all its related events.
    """
    for file_identifier, file_ in files.items():
        for event_identifier in file_["events"]:
            if event_identifier in events:
                events[event_identifier]["files"].add(file_identifier)


def relate_agents_to_events(agents, events):
    """
    Ensure the identifier of an agent is included in the events collection
    for all its related events.
    """
    for agent_identifier, agent in agents.items():
        for event_identifier in agent["events"]:
            if event_identifier in events:
                events[event_identifier]["agents"].add(agent_identifier)


def get_valid_events(files, agents, events, file_identifiers_to_ignore, printfn=print):
    """
    Get the events that are related to valid files and agents. These can
    be stored in the database.

    Return also a boolean indicating if there are incomplete (invalid) events.
    """

    result = []
    invalid_events_exist = False

    # cache the identifiers and use them for several events
    file_identifiers = set(files)
    agent_identifiers = set(agents)

    for event in events.values():
        # get the agents for this event
        event_agents = get_event_agents(event, agents, agent_identifiers, printfn)
        if not event_agents:
            log_event_without_agents(event["identifier"], printfn)
            invalid_events_exist = True
            continue

        # get the files for this event
        event_files = get_event_files(
            event, files, file_identifiers, file_identifiers_to_ignore, printfn
        )
        if not event_files:
            log_event_without_files(event["identifier"], printfn)
            invalid_events_exist = True
            continue

        result.append(
            {"event_agents": event_agents, "event_files": event_files, "event": event}
        )

    return result, invalid_events_exist


def main(job):
    # extract arguments from job
    transfer_uuid, xsd_path, xml_path = job.args[1:4]

    # a missing events XML file is not considered an error
    if not os.path.isfile(xml_path):
        log_missing_events_xml(xml_path, job.pyprint)
        return SUCCESS

    file_queryset = File.objects.filter(transfer_id=transfer_uuid)

    # validate the xml
    tree = get_xml_tree(xsd_path, xml_path, job.pyprint)
    if tree is None:
        return FAILURE

    # get dictionary of dictionaries with file attributes
    files = get_elements(
        tree, 'premis:object[@xsi:type="premis:file"]', file_element_factory
    )

    # get dictionary of dictionaries with agent attributes
    agents = get_elements(tree, "premis:agent", agent_element_factory)

    # get dictionary of dictionaries with event attributes
    events = get_elements(tree, "premis:event", event_element_factory, job.pyprint)

    # look for files that do not contain a premis:originalName element or do not
    # point to a filename in this transfer. we only care about their identifiers
    invalid_file_identifiers = get_invalid_file_identifiers(
        files, file_queryset, job.pyprint
    )

    # iterate files and relate them to their events
    relate_files_to_events(files, events)

    # iterate agents and relate them to their events
    relate_agents_to_events(agents, events)

    # log elements that are not related to any other element
    unrelated_elements_exist = print_unrelated_elements(
        files, agents, events, job.pyprint
    )

    # log elements that are related to nonexistent elements
    invalid_relationships_exist = print_nonexistent_references(
        files, agents, events, job.pyprint
    )

    # get events that can be saved
    valid_events, invalid_events_exist = get_valid_events(
        files, agents, events, invalid_file_identifiers, job.pyprint
    )

    if any(
        [unrelated_elements_exist, invalid_relationships_exist, invalid_events_exist]
    ):
        return FAILURE

    # save events
    save_events(valid_events, file_queryset, job.pyprint)

    return SUCCESS


def call(jobs):
    for job in jobs:
        with job.JobContext(logger=logger):
            with transaction.atomic():
                job.set_status(main(job))



import json
import sys

from django.db import transaction

from a3m.main import models

FIELDS = (
    "title",
    "creator",
    "subject",
    "description",
    "publisher",
    "contributor",
    "date",
    "type",
    "format",
    "identifier",
    "source",
    "relation",
    "language",
    "coverage",
    "rights",
)


def main(job, transfer_uuid, target_path):
    jsonified = {}
    try:
        dc = models.DublinCore.objects.get(metadataappliestoidentifier=transfer_uuid)
    except:  # There may not be any DC metadata for this transfer, and that's fine
        job.pyprint("No DC metadata found; skipping", file=sys.stderr)
        return 0
    for field in FIELDS:
        attr = getattr(dc, field)
        if attr:
            jsonified[field] = attr

    job.pyprint("Saving the following properties to:", target_path)
    job.pyprint(jsonified)

    with open(target_path, "w") as json_file:
        json.dump(jsonified, json_file)
    return 0


def call(jobs):
    with transaction.atomic():
        for job in jobs:
            with job.JobContext():
                transfer_uuid = job.args[1]
                target_path = job.args[2]
                job.set_status(main(job, transfer_uuid, target_path))



"""
Database related functionality.

This module implements an `auto_close_old_connections` decorator/context manager.
In order to be able to re-use database connections in the Django ORM outside of
the typical request/response cycle, we have to wrap database usage in order to
make sure anything older than CONN_MAX_AGE is cleared.

This is done via the function decorator `auto_close_old_connections()` in most
cases. For some cases (e.g. generator functions) it may be preferable to use it
as a context manager. Note that because of multiple entry points we sometimes
end up entering multiple times; that causes a slight increase in overhead, but
it is much preferred to having an unwrapped function, which may cause errors.

To debug connection timeouts, turn DEBUG on in Django settings, which will log
all SQL queries and allow us to check that all logged queries occur within the
wrapper. Note though, this will result in _very_ verbose logs.
"""
import logging
import threading
import traceback
from contextlib import ContextDecorator
from importlib import import_module
from typing import Any

from django.apps import apps
from django.conf import settings
from django.core.management.sql import emit_post_migrate_signal
from django.core.management.sql import emit_pre_migrate_signal
from django.db import close_old_connections
from django.db import connections
from django.db import DEFAULT_DB_ALIAS
from django.db.migrations.executor import MigrationExecutor
from django.db.migrations.state import ModelState
from django.utils.module_loading import module_has_submodule


logger = logging.getLogger(__name__)
thread_locals = threading.local()


class AutoCloseOldConnections(ContextDecorator):
    """
    Decorator to ensure that db connections older than CONN_MAX_AGE are
    closed before execution. Normally, one would close connections after
    execution, but we have some jobs that could run for hours out of
    process, then trigger db access.
    """

    def __enter__(self):
        close_old_connections()
        return self

    def __exit__(self, *exc):
        return False


class DebugAutoCloseOldConnections(AutoCloseOldConnections):
    """
    Debug version of AutoCloseOldConnections; logs warnings with stack traces
    to identify functions that should be wrapped.
    """

    def __enter__(self):
        if not hasattr(thread_locals, "auto_close_connections_depth"):
            thread_locals.auto_close_connections_depth = 0
        thread_locals.auto_close_connections_depth += 1
        logger.debug(
            "Entered auto close connections (depth %s)",
            thread_locals.auto_close_connections_depth,
        )
        return super().__enter__()

    def __exit__(self, *exc):
        thread_locals.auto_close_connections_depth -= 1
        logger.debug(
            "Exited auto close connections (depth %s)",
            thread_locals.auto_close_connections_depth,
        )
        return False


class CheckCloseConnectionsHandler(logging.Handler):
    """A logger that issues warnings when the database is accessed outside
    of an auto_close_old_connections wrapper.
    """

    def emit(self, record):
        if "Entered auto close connections" not in record.getMessage():
            return
        depth = getattr(thread_locals, "auto_close_connections_depth", 0)
        if depth <= 0:
            logging.getLogger("a3m.server").warning(
                "Database access occurred outside of an "
                "auto_close_old_connections context. Depth %d. Traceback %s",
                depth,
                "\n".join(traceback.format_stack()),
            )


def migration_progress_callback(action, migration=None, fake=False):
    if action == "apply_start":
        logger.info("Applying migration %s...", migration)
    elif action == "unapply_start":
        logger.info("Unapplying %s...", migration)
    elif action == "render_start":
        logger.info("Rendering model states...")


def migrate():
    """Migrate the database.

    This is a simplified version of Django's ``migrate`` management command
    that does not require ``management.call_command`` which is not functional
    in the native binary environment.
    """
    # Import the 'management' module within each installed app, to register
    # dispatcher events.
    verbosity = int(settings.DEBUG)
    interactive = False
    for app_config in apps.get_app_configs():
        if module_has_submodule(app_config.module, "management"):
            import_module(".management", app_config.name)

    connection = connections[DEFAULT_DB_ALIAS]
    connection.prepare_database()
    executor = MigrationExecutor(connection, migration_progress_callback)
    executor.loader.check_consistent_history(connection)
    conflicts = executor.loader.detect_conflicts()
    if conflicts:
        raise Exception("Conflicting mgirations detected.")
    targets = executor.loader.graph.leaf_nodes()
    plan = executor.migration_plan(targets)
    pre_migrate_state = executor._create_project_state(with_applied_migrations=True)
    pre_migrate_apps = pre_migrate_state.apps
    emit_pre_migrate_signal(
        verbosity,
        False,
        connection.alias,
        apps=pre_migrate_apps,
        plan=plan,
    )
    post_migrate_state = executor.migrate(
        targets,
        plan=plan,
        state=pre_migrate_state.clone(),
        fake=False,
        fake_initial=False,
    )
    post_migrate_state.clear_delayed_apps_cache()
    post_migrate_apps = post_migrate_state.apps
    with post_migrate_apps.bulk_update():
        model_keys = []
        for model_state in post_migrate_apps.real_models:
            model_key = model_state.app_label, model_state.name_lower
            model_keys.append(model_key)
            post_migrate_apps.unregister_model(*model_key)
    post_migrate_apps.render_multiple(
        [ModelState.from_model(apps.get_model(*model)) for model in model_keys]
    )
    emit_post_migrate_signal(
        verbosity,
        interactive,
        connection.alias,
        apps=post_migrate_apps,
        plan=plan,
    )
    logger.info("Database configured.")


auto_close_old_connections: type[Any]
if settings.DEBUG:
    logger.debug("Using DEBUG auto_close_old_connections")
    auto_close_old_connections = DebugAutoCloseOldConnections

    # Queries are only logged if DEBUG is on.
    db_logger = logging.getLogger(__name__)
    db_logger.setLevel(logging.DEBUG)
    handler = CheckCloseConnectionsHandler(level=logging.DEBUG)
    db_logger.addHandler(handler)
else:
    auto_close_old_connections = AutoCloseOldConnections


__all__ = ("auto_close_old_connections",)



# This file is part of Archivematica.
#
# Copyright 2010-2017 Artefactual Systems Inc. <http://artefactual.com>
#
# Archivematica is free software: you can redistribute it and/or modify
# it under the terms of the GNU Affero General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# Archivematica is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with Archivematica.  If not, see <http://www.gnu.org/licenses/>.
"""Verify Checksum Job

Archivematica wraps the hashsum utility to verify checksums provided to the
system as part of a transfer, e.g. checksum.md5 in the transfer metadata
folder. We wrap this coreutils utility by using a Hashsum class which enables
us to easily call each of the tools packaged against its different algorithms:

    * MD5
    * SHA1
    * SHA256
    * SHA512
"""
import datetime
import logging
import os
import shlex
import subprocess  # nosec B404
import sys
import uuid

from django.db import transaction

from a3m.archivematicaFunctions import strToUnicode
from a3m.main.models import Event
from a3m.main.models import File
from a3m.main.models import Transfer


logger = logging.getLogger(__name__)


class NoHashCommandAvailable(Exception):
    """Provide feedback to the user if the checksum command cannot be found
    for the provided checksum file.
    """


class PREMISFailure(Exception):
    """Provide feedback to the user if there is a problem writing PREMIS event
    information to the database.
    """


class Hashsum:
    """Class to capture various functions around calling Hashsum as a mechanism
    for comparing user-supplied checksums in Archivematica.
    """

    # Key-values consisting of a "hash file" specific to a checksum algorithm,
    # and the hashsum command that we want to call against the file.
    HASHFILES_COMMANDS = {
        "checksum.md5": "md5sum",
        "checksum.sha1": "sha1sum",
        "checksum.sha256": "sha256sum",
        "checksum.sha512": "sha512sum",
    }

    OKAY_STRING = ": OK"
    FAIL_STRING = ": FAILED"
    ZERO_STRING = "no properly formatted"
    IMPROPER_STRING = "improperly formatted"
    FAILED_OPEN = ": FAILED open or read"
    EXIT_NON_ZERO = "returned non-zero exit status 1"

    def __init__(self, path, job=print):
        """Initialize an object prepared to invoke a hashsum command on a given
        hash file at a given path and return properties related to that
        invocation to the caller, e.g. to add provenance information to PREMIS
        output.
        """
        try:
            self.job = job
            self.command_called = None
            self.hashfile = path
            self.COMMAND = self.HASHFILES_COMMANDS[os.path.basename(path)]
        except KeyError:
            raise NoHashCommandAvailable()

    def _call(self, *args, **kwargs):
        """Make the call to Python subprocess and record the command being
        called.
        """
        self.command_called = shlex.split(" ".join((self.COMMAND,) + args))
        return self._decode(
            subprocess.check_output(
                self.command_called,  # noqa S603
                cwd=kwargs.get("transfer_dir"),
            )
        )

    def count_and_compare_lines(self, objects_dir):
        """Count the number of lines in a checksum file and compare with the
        number of objects being transferred. The requirement of hashsum as
        this microservice job is written is that the mapping is 1:1. There
        isn't space for an empty line at the end of the file.
        """
        lines = self._count_lines(self.hashfile)
        objects = self._count_files(objects_dir)
        if lines == objects:
            return True
        self.job.pyprint(
            "{}: Comparison failed with {} checksum lines and {} "
            "transfer files".format(self.get_ext(self.hashfile), lines, objects),
            file=sys.stderr,
        )
        return False

    def compare_hashes(self, transfer_dir):
        """Compare transfer files with the checksum file provided."""
        objects_dir = os.path.join(transfer_dir, "objects")
        if not self.count_and_compare_lines(objects_dir):
            return 1
        try:
            self._call("-c", "--strict", self.hashfile, transfer_dir=objects_dir)
            return 0
        except subprocess.CalledProcessError as err:
            if self.EXIT_NON_ZERO in str(err):
                warn = "{}: comparison exited with status: {}. Please check the formatting of the checksums or integrity of the files.".format(
                    self.get_ext(self.hashfile), err.returncode
                )
                self.job.pyprint(warn, file=sys.stderr)
            for line in self._decode(err.output):
                if not line:
                    continue
                if line.endswith(self.OKAY_STRING):
                    continue
                if (
                    line.endswith(self.FAIL_STRING)
                    or self.ZERO_STRING in line
                    or self.IMPROPER_STRING in line
                ):
                    self.job.pyprint(
                        f"{self.get_ext(self.hashfile)}: {line}",
                        file=sys.stderr,
                    )
                if line.endswith(self.FAILED_OPEN):
                    self.job.pyprint(
                        f"{self.get_ext(self.hashfile)}: {line}",
                        file=sys.stderr,
                    )
            return err.returncode

    def version(self):
        """Return version information for the command being called."""
        try:
            return self._call("--version")[0]
        except subprocess.CalledProcessError:
            return self.COMMAND

    def get_command_detail(self):
        """Provide some way for the user to get information out of the class to
        write METS/PREMIS provenance information.
        """
        if not self.command_called:
            err_str = "Unable to retrieve information about the hashsum command called by the script"
            raise PREMISFailure(err_str)
        return 'program="{}"; version="{}"'.format(
            " ".join(self.command_called), self.version()
        )

    @staticmethod
    def _decode(out):
        """Decode binary string and return a list of lines."""
        return str(out, "utf8").split("\n")

    @staticmethod
    def get_ext(path):
        """Return the extension of the checksum file provided"""
        ext = os.path.splitext(path)[1]
        if not ext:
            return path
        return ext.replace(".", "")

    @staticmethod
    def _count_lines(path):
        """Count the number of lines in a checksum file."""
        count = 0
        with open(path) as hashfile:
            for count, _ in enumerate(hashfile):
                pass
        # Negate zero-based count.
        return count + 1

    @staticmethod
    def _count_files(path):
        """Walk the directories on a given path and count the number of files."""
        return sum(len(files) for _, _, files in os.walk(path))


def get_file_queryset(transfer_uuid):
    """
    Return a queryset for File objects related to the SIP/Transfer.
    """
    file_objs_queryset = File.objects.filter(
        **{"removedtime__isnull": True, "transfer_id": transfer_uuid}
    )
    if not file_objs_queryset.exists():
        err_str = "Unable to find the transfer objects for the SIP: '{}' in the database".format(
            transfer_uuid
        )
        raise PREMISFailure(err_str)
    return file_objs_queryset


def write_premis_event_per_file(file_uuids, transfer_uuid, event_detail):
    """Generate PREMIS events per File object verified in this transfer."""
    event_type = "fixity check"
    event_outcome = "pass"
    events = []
    agents = Transfer.objects.get(uuid=transfer_uuid).agents
    with transaction.atomic():
        for file_obj in file_uuids:
            checksum_event = Event(
                file_uuid=file_obj,
                event_type=event_type,
                event_datetime=datetime.datetime.now(),
                event_outcome=event_outcome,
                event_detail=event_detail,
                event_id=uuid.uuid4(),
            )
            events.append(checksum_event)
        # All the events sit in memory at this point and are then written.
        # We could write this in batches if we need to optimize further.
        Event.objects.bulk_create(events)
        # Adding many-to-many fields with bulk create is awkward, we have to
        # loop through again.
        for event in Event.objects.filter(
            file_uuid__in=[event.file_uuid for event in events]
        ):
            event.agents.add(*agents)


def run_hashsum_commands(job):
    """Run hashsum commands and generate a cumulative return code."""
    transfer_dir = None
    transfer_uuid = None
    try:
        transfer_dir = strToUnicode(job.args[1])
        transfer_uuid = job.args[2]
    except IndexError:
        logger.error("Cannot access expected module arguments: %s", job.args)
        return 1
    ret = 0
    # Create a query-set once so we don't need to generate per each checksum
    # file type.
    file_queryset = get_file_queryset(transfer_uuid)
    for hashfile in Hashsum.HASHFILES_COMMANDS:
        hashsum = None
        hashfilepath = os.path.join(transfer_dir, "metadata", hashfile)
        if os.path.exists(hashfilepath):
            try:
                hashsum = Hashsum(hashfilepath, job)
            except NoHashCommandAvailable:
                job.pyprint(
                    "Nothing to do for {}. No command available.".format(
                        Hashsum.get_ext(hashfilepath)
                    )
                )
                continue
        if hashsum:
            job.pyprint(
                "Comparing transfer checksums with the supplied {} file".format(
                    Hashsum.get_ext(hashfilepath)
                ),
                file=sys.stderr,
            )
            result = hashsum.compare_hashes(transfer_dir=transfer_dir)
            # Add to PREMIS on success only.
            if result == 0:
                job.pyprint(f"{Hashsum.get_ext(hashfile)}: Comparison was OK")
                write_premis_event_per_file(
                    file_uuids=file_queryset,
                    transfer_uuid=transfer_uuid,
                    event_detail=hashsum.get_command_detail(),
                )
                continue
            ret += result
    return ret


def call(jobs):
    """Primary entry point for MCP Client script."""
    for job in jobs:
        with job.JobContext(logger=logger):
            job.set_status(run_hashsum_commands(job))



import os
import pathlib


THIS_DIR = os.path.dirname(__file__)

ASSETS_DIR = os.path.join(THIS_DIR, "assets", "")

# Use local XML schemas for validation.
os.environ["XML_CATALOG_FILES"] = str(
    pathlib.Path(__file__).parent / "assets/catalog/catalog.xml"
)



import logging
from collections.abc import Callable

import tenacity
from grpc import Channel
from grpc import RpcError

from a3m import __version__
from a3m.api.transferservice import v1beta1 as transfer_service_api


logger = logging.getLogger(__name__)


# Default duration in seconds of RPC calls.
_GRPC_DEFAULT_TIMEOUT_SECS = 30

# Metadata key containing the client version.
_VERSION_METADATA_KEY = "version"


class Client:
    """a3m gRPC API client."""

    def __init__(
        self,
        channel: Channel,
        rpc_timeout: int | None = _GRPC_DEFAULT_TIMEOUT_SECS,
        wait_for_ready: bool = False,
    ):
        self.transfer_stub = transfer_service_api.service_pb2_grpc.TransferServiceStub(
            channel
        )
        self.rpc_timeout = rpc_timeout
        self.wait_for_ready = wait_for_ready

    def _unary_call(self, api_method, request):
        rpc_name = request.__class__.__name__.replace("Request", "")
        logger.debug("RPC call %s with request: %r", rpc_name, request)
        try:
            return api_method(
                request,
                timeout=self.rpc_timeout,
                metadata=Client.version_metadata(),
                wait_for_ready=self.wait_for_ready,
            )
        except RpcError as e:
            logger.warning("RPC call %s got error %s", rpc_name, e)
            raise

    @staticmethod
    def version_metadata():
        return ((_VERSION_METADATA_KEY, __version__),)

    def submit(
        self,
        url: str,
        name: str,
        config: transfer_service_api.request_response_pb2.ProcessingConfig = None,
    ):
        request = transfer_service_api.request_response_pb2.SubmitRequest(
            name=name, url=url, config=config
        )
        return self._unary_call(self.transfer_stub.Submit, request)

    def read(self, package_id: str):
        request = transfer_service_api.request_response_pb2.ReadRequest(id=package_id)
        return self._unary_call(self.transfer_stub.Read, request)

    def wait_until_complete(
        self, package_id: str, spin_cb: Callable = None
    ) -> transfer_service_api.request_response_pb2.ReadResponse:
        """Blocks until processing of a package has completed."""

        def _should_continue(
            resp: transfer_service_api.request_response_pb2.ReadResponse,
        ):
            return (
                resp.status
                == transfer_service_api.request_response_pb2.PACKAGE_STATUS_PROCESSING
            )

        def _callback(retry_state):
            if spin_cb is not None:
                spin_cb(retry_state)

        @tenacity.retry(
            wait=tenacity.wait_fixed(1),
            retry=tenacity.retry_if_result(_should_continue),
            after=_callback,
        )
        def _poll():
            """Retries while the package is processing."""
            return self.read(package_id)

        return _poll()

    def list_tasks(self, job_id: str):
        request = transfer_service_api.request_response_pb2.ListTasksRequest(
            job_id=job_id
        )
        return self._unary_call(self.transfer_stub.ListTasks, request)



"""Runs zero or more FPR validation commands against the provided file and
returns an exit code. May also print to stdout, generate an Event model in the
db, and/or write command-specific stdout to disk.

If a format has no defined validation commands, no command is run.

Arguments:
    [FILE_PATH] [FILE_UUID] [SIP_UUID] [SHARED_PATH] [FILE_TYPE]

"""
import ast
import logging
import os
import sys
from pprint import pformat
from typing import Final

from django.db import transaction

from a3m import databaseFunctions
from a3m.dicts import replace_string_values
from a3m.dicts import setup_dicts
from a3m.executeOrRunSubProcess import executeOrRun
from a3m.fpr.registry import CommandScriptType
from a3m.fpr.registry import FPR
from a3m.fpr.registry import Rule
from a3m.fpr.registry import RulePurpose
from a3m.main.models import Derivation
from a3m.main.models import File
from a3m.main.models import SIP


SUCCESS_CODE = 0
FAIL_CODE = 1
NOT_DERIVATIVE_CODE = 0
NO_RULES_CODE = 0
DERIVATIVE_TYPES = ("preservation", "access")


def main(job, file_path, file_uuid, sip_uuid, shared_path, file_type):
    setup_dicts()

    validator = Validator(job, file_path, file_uuid, sip_uuid, shared_path, file_type)
    return validator.validate()


logger = logging.getLogger(__name__)


class Validator:
    """A validator validates a file, during transfer or ingest, using FPR
    config.

    If the file is an access derivative or a preservation derivative, the
    ``file_type`` will be 'access' or 'preservation', respectively. This will
    trigger specialized behaviour. The default (an original) behaviour
    corresponds to a ``file_type`` of 'original'.

    Usage: initialize on a file and then call the ``validate`` method to
    determine whether a given file conforms to a given specification.
    """

    purpose: Final[RulePurpose] = RulePurpose.VALIDATION

    def __init__(self, job, file_path, file_uuid, sip_uuid, shared_path, file_type):
        self.job = job
        self.file_path = file_path
        self.file_uuid = file_uuid
        self.sip_uuid = sip_uuid
        self.shared_path = shared_path
        self.file_type = file_type
        self._sip_logs_dir = None
        self._sip_pres_val_dir = None

    def validate(self):
        """Validate the file identified by ``self.file_uuid``, using all rules
        that apply. Return an error code (1 or 0), which the script as a whole
        also returns. Side effects include printing to stdout/stderr (which
        gets displayed in the dashboard), creating Event models in the db, and
        saving log/stdout files within the unit's (Transfer/SIP) directory
        structure.
        """
        if self.file_type in DERIVATIVE_TYPES and not self._file_is_derivative():
            self.job.print_output(
                "File {uuid} {not_derivative_msg}; not validating.".format(
                    uuid=self.file_uuid, not_derivative_msg=self._not_derivative_msg()
                )
            )
            return NOT_DERIVATIVE_CODE
        rules = self._get_rules()
        if not rules:
            return NO_RULES_CODE
        rule_outputs = []
        for rule in rules:
            rule_outputs.append(self._execute_rule_command(rule))
        if "failed" in rule_outputs:
            return FAIL_CODE
        return SUCCESS_CODE

    def _get_rules(self):
        """Return all FPR rules that apply to files of this type."""
        return FPR.get_file_rules(self.file_uuid, self.purpose, fallback=True)

    def _execute_rule_command(self, rule: Rule):
        """Run the command against the file and return either 'passed' or
        'failed'. If the command errors or determines that the file is invalid,
        return 'failed'. Non-errors will result in the creation of an Event
        model in the db. Preservation derivative validation will result in the
        stdout from the command being saved to disk within the unit (i.e., SIP).
        """
        result = "passed"
        if rule.command.script_type in (
            CommandScriptType.BASH_SCRIPT,
            CommandScriptType.COMMAND,
        ):
            command_to_execute = replace_string_values(
                rule.command.command,
                file_=self.file_uuid,
                sip=self.sip_uuid,
                type_="file",
            )
            args = []
        else:
            command_to_execute = rule.command.command
            args = [self.file_path]
        self.job.print_output("Running", rule.command.description)
        exitstatus, stdout, stderr = executeOrRun(
            type=rule.command.script_type.value,
            command=command_to_execute,
            printing=False,
            arguments=args,
        )
        if exitstatus != 0:
            self.job.print_error(
                "Command {description} failed with exit status {status};"
                " stderr: {stderr}".format(
                    description=rule.command.description,
                    status=exitstatus,
                    stderr=stderr,
                )
            )
            return "failed"
        # Parse output and generate an Event
        # TODO: Evaluating a python string from a user-definable script seems
        # insecure practice; should be JSON.
        output = ast.literal_eval(stdout)
        event_detail = 'program="{tool.description}"; version="{tool.version}"'.format(
            tool=rule.command.tool
        )
        # If the FPR command has not errored but the actual validation
        # determined that the file is not valid, then we want to both create a
        # validation event in the db and set ``failed`` to ``True`` because we
        # want the micro-service in the dashboard GUI to indicate "Failed".
        # NOTE: this requires that the stdout of all validation FPR commands be
        # a dict (preferably a JSON object) with an ``eventOutcomeInformation``
        # boolean attribute.
        if output.get("eventOutcomeInformation") == "pass":
            self.job.print_output(
                f'Command "{rule.command.description}" was successful'
            )
        elif output.get("eventOutcomeInformation") == "partial pass":
            self.job.print_output(
                f'Command "{rule.command.description}" was partially successful'
            )
        else:
            self.job.pyprint(
                "Command {cmd_description} indicated failure with this"
                " output:\n\n{output}".format(
                    cmd_description=rule.command.description, output=pformat(stdout)
                ),
                file=sys.stderr,
            )
            result = "failed"
        if self.file_type == "preservation":
            self._save_stdout_to_logs_dir(output)
        self.job.print_output(
            "Creating {purpose} event for {file_path} ({file_uuid})".format(
                purpose=self.purpose, file_path=self.file_path, file_uuid=self.file_uuid
            )
        )
        databaseFunctions.insertIntoEvents(
            fileUUID=self.file_uuid,
            eventType="validation",  # From PREMIS controlled vocab.
            eventDetail=event_detail,
            eventOutcome=output.get("eventOutcomeInformation"),
            eventOutcomeDetailNote=output.get("eventOutcomeDetailNote"),
        )
        return result

    def _save_stdout_to_logs_dir(self, output):
        """Save the validation command's output from validating the file to a
        file at logs/implementationChecks/<input_filename>.xml in the SIP.
        ``output`` is expected to be a dict with a ``stdout`` key.
        """
        stdout = output.get("stdout")
        if stdout and self.sip_pres_val_dir:
            filename = os.path.basename(self.file_path)
            stdout_path = os.path.join(self.sip_pres_val_dir, f"{filename}.xml")
            with open(stdout_path, "w") as f:
                f.write(stdout)

    def _file_is_derivative(self):
        """Return ``True`` if the file we are validating is a derivative, i.e.,
        a modified version created for preservation or access.
        """
        if self.file_type == "preservation":
            return self._file_is_preservation_derivative()
        return self._file_is_access_derivative()

    def _file_is_preservation_derivative(self):
        """Returns ``True`` if the file with UUID ``self.file_uuid`` is a
        preservation derivative.
        """
        try:
            Derivation.objects.get(
                derived_file__uuid=self.file_uuid, event__event_type="normalization"
            )
            return True
        except Derivation.DoesNotExist:
            return False

    def _file_is_access_derivative(self):
        """Returns ``True`` if the file with UUID ``self.file_uuid`` is an
        access derivative.
        """
        try:
            file_model = File.objects.get(uuid=self.file_uuid)
            if file_model.filegrpuse == "access":
                try:
                    Derivation.objects.get(
                        derived_file__uuid=self.file_uuid, event__isnull=True
                    )
                    return True
                except Derivation.DoesNotExist:
                    return False
            else:
                return False
        except File.DoesNotExist:
            return False

    def _not_derivative_msg(self):
        """Return the message to print if the file is not a derivative."""
        if self.file_type == "preservation":
            return "is not a preservation derivative"
        return "is not an access derivative"

    @property
    def sip_logs_dir(self):
        """Return the absolute path the logs/ directory of the SIP that the
        target file is a part of.
        """
        if self._sip_logs_dir:
            return self._sip_logs_dir
        try:
            sip_model = SIP.objects.get(uuid=self.sip_uuid)
        except (SIP.DoesNotExist, SIP.MultipleObjectsReturned):
            self.job.print_error(
                "Warning: unable to retrieve SIP model corresponding to SIP"
                " UUID {}".format(self.sip_uuid)
            )
            return None
        else:
            sip_path = sip_model.currentpath.replace(
                "%sharedPath%", self.shared_path, 1
            )
            logs_dir = os.path.join(sip_path, "logs")
            if os.path.isdir(logs_dir):
                self._sip_logs_dir = logs_dir
                return logs_dir
            self.job.print_error(
                "Warning: unable to find a logs/ directory in the SIP"
                " with UUID {}".format(self.sip_uuid)
            )
            return None

    @property
    def sip_pres_val_dir(self):
        """Return the full path to the directory within the SIP where stdout
        from perservation derivative validation output should be written to
        disk.
        TODO: the leaf dir of the path return is 'implementationChecks' because
        this functionality was implemented with MediaConch implementation
        checks in mind. However, if saving command stdout to the logs/ dir of a
        SIP is needed for validation based on other tools, this directory name
        may need to be changed in a context-sensitive way.
        """
        if self._sip_pres_val_dir:
            return self._sip_pres_val_dir
        if self.sip_logs_dir:
            _sip_pres_val_dir = os.path.join(self.sip_logs_dir, "implementationChecks")
            if os.path.isdir(_sip_pres_val_dir):
                self._sip_pres_val_dir = _sip_pres_val_dir
            else:
                try:
                    os.makedirs(_sip_pres_val_dir)
                except OSError:
                    pass
                else:
                    self._sip_pres_val_dir = _sip_pres_val_dir
        return self._sip_pres_val_dir


def _get_shared_path(argv):
    try:
        return argv[4]
    except IndexError:
        return None


def _get_file_type(argv):
    try:
        return argv[5]
    except IndexError:
        return "original"


def call(jobs):
    with transaction.atomic():
        for job in jobs:
            with job.JobContext(logger=logger):
                file_path = job.args[1]
                file_uuid = job.args[2]
                sip_uuid = job.args[3]
                shared_path = _get_shared_path(job.args)
                file_type = _get_file_type(job.args)
                job.set_status(
                    main(job, file_path, file_uuid, sip_uuid, shared_path, file_type)
                )
